import unittest
try:
    from zoneinfo import ZoneInfo  # will run only in python 3.9+
except ImportError:
    from backports.zoneinfo import ZoneInfo  # backport to python < 3.9
from wiliot_deployment_tools.common.utils import *

class UtilsTest(unittest.TestCase):

    def test_convert_datetime_to_timestamp(self):
        now = datetime.datetime.now()
        timestamp = now.timestamp()
        year = now.year
        month = now.month
        day = now.day
        hour = now.hour
        minute = now.minute
        second = now.second
        micro = now.microsecond
        self.assertEquals(convert_datetime_to_timestamp(year, month, day, hour, minute, second, micro, hours_from_utc=2), math.ceil(timestamp * 1000))

    def test_ms_timestamp_to_timezone(self):
        now = datetime.datetime.now()
        ms_timestamp = now.timestamp()*1000
        tzs = ['Israel', 'America/Los_Angeles', 'Europe/Vienna']
        for tz in tzs:
            timezone = ZoneInfo(tz)
            # hour = true, milli = true
            strftime = now.astimezone(timezone).strftime('%Y-%m-%d %H:%M:%S.%f %Z')
            self.assertEquals(mstimestamp_to_timezone(ms_timestamp, tz, milli=True, hour=True), strftime)
            # hour = true, milli = false
            strftime = now.astimezone(timezone).strftime('%Y-%m-%d %H:%M:%S %Z')
            self.assertEquals(mstimestamp_to_timezone(ms_timestamp, tz, milli=False, hour=True), strftime)
            # hour = false
            strftime = now.astimezone(timezone).strftime('%Y-%m-%d')
            self.assertEquals(mstimestamp_to_timezone(ms_timestamp, tz, hour=False), strftime)

    def test_convert_timestamp_to_datetime(self):
        now = datetime.datetime.now()
        timestamp = now.timestamp()
        ms_timestamp = timestamp*1000
        self.assertEquals(now, convert_timestamp_to_datetime(ms_timestamp))
        self.assertEquals(now, convert_timestamp_to_datetime(timestamp))
    
    def test_datetime_to_timezone(self):
        now = datetime.datetime.now()
        tzs = ['Israel', 'America/Los_Angeles', 'Europe/Vienna']
        for tz in tzs:
            self.assertEquals(now.astimezone(ZoneInfo(tz)), datetime_to_timezone(now, tz))

if __name__ == '__main__':
    unittest.main()



import datetime
import unittest
from wiliot_deployment_tools.power_mgmt.power_mgmt import PowerManagementClient, PowerManagementPacket
from api_secrets import *
class TestPowerManagement(unittest.TestCase):

    def test_init(self):
        mgmt = PowerManagementClient(WH_EDGE, WH_OWNERID)
        self.assertIsInstance(mgmt, PowerManagementClient)
        
    def test_check_gw_compatible(self):
        mgmt = PowerManagementClient(WH_EDGE, WH_OWNERID)
        self.assertTrue(mgmt.check_gw_compatible_for_pwr_mgmt(HOME_GW))
    
    def test_get_brg_relevant_gw(self):
        mgmt = PowerManagementClient(WH_EDGE, WH_OWNERID)
        self.assertEquals(mgmt.get_bridge_relevant_gw(HOME_BRG), HOME_GW)
        
    def test_get_relevant_bridges(self):
        mgmt = PowerManagementClient(WH_EDGE, WH_OWNERID)
        self.assertEquals(mgmt.get_relevant_bridges(HOME_GW), [HOME_BRG])
    
    def test_check_ble_power_mgmt_support(self):
        ble_vers = [('3.7.25', False), ('3.8.20', False), ('3.11.39', False), ('3.11.40', True), ('3.12.40', True), ('4.0.15', True), ('-1', False)]
        for ver, support in ble_vers:
            actual = PowerManagementClient.check_ble_power_mgmt_support(ver)
            self.assertEquals(actual, support, f'{ver}, {actual}, {support}')
        
    def test_power_mgmt_functionality(self):
        with self.subTest('init_client'):
            mgmt = PowerManagementClient(WH_EDGE, WH_OWNERID)
        with self.subTest('send_blank_packet'):
            now = datetime.datetime.now()
            self.assertTrue(mgmt.send_pwr_mgmt_pkt(bridge_id=DESK_BRG, packet=PowerManagementPacket.exit_packet(), gateway_id=DESK_GW))
        with self.subTest('send_until_ack'):
            self.assertEquals(mgmt.send_packet_until_ack(DESK_GW, DESK_BRG, PowerManagementPacket.exit_packet()), {DESK_BRG: True})    
        with self.subTest('exit_power_mgmt'):
            self.assertEquals(mgmt.exit_power_mgmt(DESK_GW, False, DESK_BRG, False), {DESK_BRG: True})
if __name__ == '__main__':
    unittest.main()



import datetime
import sys
import time
from wiliot_deployment_tools.api.extended_api import BridgeThroughGatewayAction, ExtendedEdgeClient, GatewayType
from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.common.utils_defines import  SEC_TO_SEND


class CalibrationManagementError(Exception):
    pass
  

class CalibrationManagementClient(ExtendedEdgeClient):


    def get_bridge_relevant_gw(self, bridge_id):
        """
        Returns relevant GW for sending actions to bridge
        :param bridge_id: Bridge ID
        :rtype: str
        :return: relevant Gateway ID
        """

        # TODO - return list of potential GWs
        potential_gws = list()
        owner_gws = self.get_gateways()
        owner_gw_ids = []
        owner_gw_ids.extend(g['gatewayId'] for g in owner_gws)
        for gw in self.get_bridge(bridge_id)['connections']:
            if gw['gatewayId'] not in owner_gw_ids:
                continue
            gw_id = gw['gatewayId']
            gw_dict = self.get_gateway(gw_id)
            if gw['connected'] and \
                    bridge_id in self.get_seen_bridges(gw_id) and \
                    self.check_gw_compatible_for_action(gw_id):
                if bridge_id in self.get_seen_bridges(gw['gatewayId']):
                    potential_gws.append(gw['gatewayId'])
        if len(potential_gws) == 0:
            raise CalibrationManagementError(f'No relevant GW connected to bridge! Check deployment')
        return potential_gws[0]

    def get_relevant_bridges(self, gateway_id):
        """
        Returns seen bridges by gateway that are associated to owner and have ble version with calibration management support
        :type gateway_id: str
        :param gateway_id: gateway ID
        :rtype: list
        :return: list of relevant bridge IDs
        """
        bridges = self.get_seen_bridges(gateway_id)
        relevant_bridges = bridges
        for bridge in bridges:
            brg_ble = self.get_brg_ble_version(bridge)
            if not self.check_ble_calibration_mgmt_support(brg_ble):
                # debug_print(f'Bridge {bridge} does not support calibration management!')
                relevant_bridges.remove(bridge)
        return relevant_bridges


    @staticmethod
    def check_ble_calibration_mgmt_support(ble_ver):
        """
        checks if calibration management is supported for BLE version
        :rtype: bool
        :return: True if supported, False otherwise
        """
        ble_ver = ble_ver.split('.')
        if not ((int(ble_ver[0]) > 3) or
                (int(ble_ver[0]) > 2 and int(ble_ver[1]) > 14) or \
                (int(ble_ver[0]) > 2 and int(ble_ver[1]) > 13 and int(ble_ver[2]) > 61)):
            return False
        else:
            return True


    def send_calib_mgmt_pkt(self, bridge_id, payload, gateway_id=None,
                          silent=False, seconds_to_send=1):
        """
        send calibration management packet to bridge
        :type bridge_id: str
        :param bridge_id: Bridge ID
        :type packet: CalibrationManagementPacket
        :param packet: packet to send to bridge
        :type gateway_id: str
        :param gateway_id: Gateway ID
        :type silent: bool
        :param silent: print calibration management packet parameters
        :type seconds_to_send: int
        :param seconds_to_send: seconds to repeat packet
        :rtype: bool
        :return: True if sent, False otherwise
        """
        if gateway_id is None:
            via_gw = self.get_bridge_relevant_gw(bridge_id)
        else:
            via_gw = gateway_id
        brg_version = None
        brg_version = self.get_brg_ble_version(bridge_id)
        gw_version = self.get_gw_ble_version(via_gw)
        for version in filter(None, [brg_version, gw_version]):
            if not self.check_ble_calibration_mgmt_support(version):
                return False
        rxtx = datetime.timedelta(milliseconds=
                                  self.get_gateway(via_gw)['reportedConf']['additional']['rxTxPeriodMs'])
        seconds_to_send = datetime.timedelta(seconds=seconds_to_send)
        reps = seconds_to_send // rxtx + 1
        if not silent:
            debug_print(f'Sending packet through GW {via_gw} for {seconds_to_send.total_seconds()} seconds, rxTx = {int(rxtx.total_seconds()*1000)} ms, reps = {reps}')
        res = self.send_bridge_action_through_gw(gateway_id,
                                                 BridgeThroughGatewayAction.SPARSE_37,
                                                 payload, bridge_id, reps=reps)
        return res

    
    def check_calib_mgmt_received(self, gateway_id, bridge_id, payload, time_sent):
        """
        check if calibration management packet was received by bridge
        :type gateway_id: str
        :param gateway_id: Gateway ID
        :type bridge_id: str
        :param bridge_id: Bridge ID
        :type packet: CalibrationManagementPacket
        :param packet: packet to check
        :type time_sent: datetime
        :param time_sent: time packet was sent
        :rtype bool
        :return: True if packet was received, None otherwise
        """
        hours_back = (datetime.datetime.now() - time_sent + datetime.timedelta(seconds=5)).total_seconds() / 3600
        acks = self.get_acks(gateway_id, hours_back)
        if acks is None:
            return None
        acks = acks[(acks['actionType'] == BridgeThroughGatewayAction.SPARSE_37) &
                    (acks['bridgeId'] == bridge_id) &
                    (acks['payload'].apply(lambda x: x[1])
                     == payload[1])] 
        if len(acks) > 0:
            return acks.to_dict()
        return None

   
    def send_calibration_packet_until_ack(self, gateway_id, bridge_id, minutes_timeout=3, mode=0):
        """
        send calibration management packet repeatedly until it is received by the bridge
        :type gateway_id: str
        :param gateway_id: Gateway ID
        :type bridge_id: str
        :param bridge_id: Bridge ID
        :type packet: CalibrationManagementPacket
        :param packet: packet to send
        :type minutes_timeout: int
        :param minutes_timeout: timeout
        :rtype: dict
        :return: dict of {bridgeID: ack status (bool)} pairs
        """
        time_sent = datetime.datetime.now()
        payload = '0{:01d}{:019d}'.format(mode, 0)
        debug_print(f'Configuring {bridge_id} To Calibration mode {mode} | Max Waiting time {minutes_timeout} minutes')
        self.send_calib_mgmt_pkt(bridge_id, payload, gateway_id, silent=False, seconds_to_send=SEC_TO_SEND)
        time_gw_stops = time_sent + datetime.timedelta(seconds=SEC_TO_SEND)
        time_end = time_sent + datetime.timedelta(minutes=minutes_timeout)
        curr_time = time_sent
        ack = None
        print('Waiting For Ack...', end='', flush=True)
        while curr_time < time_end and ack is None:
            curr_time = datetime.datetime.now()
            if curr_time > time_gw_stops - datetime.timedelta(seconds=1):  # 1 second buffer
                self.send_calib_mgmt_pkt(bridge_id, payload, gateway_id, silent=True, seconds_to_send=SEC_TO_SEND)
                time_gw_stops = datetime.datetime.now() + datetime.timedelta(seconds=SEC_TO_SEND)
                print('!', end='')
                sys.stdout.flush()
            time.sleep(0.5)
            print('.', end='')
            sys.stdout.flush()
            ack = self.check_calib_mgmt_received(gateway_id, bridge_id, payload, time_sent)
        print()  # enter newline
        ack_received = False
        if ack is not None:
            ack_received = True
            debug_print(f'Ack Received after {datetime.datetime.now() - time_sent}')
        else:
            debug_print(f'Ack not received after {datetime.datetime.now() - time_sent}!')
        return {bridge_id: ack_received}


import pprint
from wiliot_deployment_tools.api.extended_api import AndroidGatewayAction, ExtendedEdgeClient
from argparse import ArgumentParser
from wiliot_core import check_user_config_is_ok
import colorama

def main():
    parser = ArgumentParser(prog='wlt-log',
                            description='Log Viewer - CLI Tool to view Wiliot Gateway logs')
    required = parser.add_argument_group('required arguments')
    required.add_argument('-owner', type=str, help="Owner ID", required=True)
    required.add_argument('-gw', type=str, help="Gateway ID", required=True)
    parser.add_argument('-test', action='store_true',
                    help='If flag used, use test environment (prod is used by default)')
    parser.add_argument('--mobile_ble_logging', choices=['on', 'off'], help='set mobile BLE logging to on or off')
    
    args = parser.parse_args()
    if args.test:
        env = 'test'
    else:
        env = 'prod'

    owner_id = args.owner
    conf_env = env if env == 'prod' else 'non-prod'
    user_config_file_path, api_key, is_success = check_user_config_is_ok(owner_id, conf_env, 'edge')
    if is_success:
        print('credentials saved/upload from {}'.format(user_config_file_path))
    else:
        raise Exception('invalid credentials - please try again to login')

    colorama.init()

    e = ExtendedEdgeClient(api_key, args.owner, env)
    if args.mobile_ble_logging is not None:
        print(colorama.Fore.LIGHTBLACK_EX, end=None)
        print(colorama.Fore.GREEN +
          f'---Setting mobile BLE Logging {args.gw} to {args.mobile_ble_logging}---' + colorama.Style.RESET_ALL)        
        mode = {'on': AndroidGatewayAction.ENABLE_BLE_LOGS, 'off': AndroidGatewayAction.DISABLE_BLE_LOGS}
        res = e.send_action_to_gateway(args.gw, mode[args.mobile_ble_logging])
        print({args.gw: res})
    else:
        print(colorama.Fore.LIGHTBLACK_EX, end=None)
        print(colorama.Fore.GREEN +
            f'---Printing info for {args.gw}---' + colorama.Style.RESET_ALL)
        print(colorama.Style.RESET_ALL)
        try:
            print_gw(e, args.gw)
        except AttributeError as e:
            print(e)

def print_gw(e, gw):
    info = e.get_gateway_info(gw)

    print(colorama.Fore.GREEN +
        f'---Gateway Info---' + colorama.Style.RESET_ALL)
    pprint.pprint(info)
    print()
    print(colorama.Fore.GREEN +
        f'---Gateway Logs---' + colorama.Style.RESET_ALL)
    e.print_gateway_logs(gw)
    
def main_cli():
    main()


if __name__ == '__main__':
    main()



# External libraries
import math
import pandas as pd
import numpy as np
import datetime
from collections import defaultdict


from wiliot_deployment_tools.common.debug import is_databricks

try:
    from zoneinfo import ZoneInfo  # will run only in python 3.9+
except ImportError:
    from backports.zoneinfo import ZoneInfo  # backport to python < 3.9

# CSV Parsing
def get_ref_tags(csv_filepath):
    # TODO - fix to insert header if not given
    """
        returns an array of tags externalIds from a csv including one column of all the tags
        :type csv_filepath: string
        :param csv_filepath: filepath to csv with one column including all the tags
        output type: array
        output param: array of tags externalIds from a csv
    """
    tags_list = pd.read_csv(csv_filepath)
    tags_list = tags_list.to_dict()
    res = defaultdict(list)
    for key in tags_list.keys():
        for idx in tags_list[key]:
            if type(tags_list[key][idx]) == str and len(tags_list[key][idx]) == 31:
                res[key].append(tags_list[key][idx])
    return res

def get_bridges_from_csv(csv_filepath):
    """
        :type csv_filepath: string
        :param csv_filepath: bridges names csv file path
    """
    if csv_filepath is None:
        return None
    bridges_df = pd.read_csv(csv_filepath)
    zones_dict = bridges_df.to_dict('list')
    return zones_dict

# Date & Time Related
def convert_datetime_to_timestamp(year=2022, month=1, day=1, hour=0, minute=0, seconds=0, micro_secs=0,
                                hours_from_utc=0):
    """
    returns the timestamp of Israeli datetime
    :type year: int
    :param year: year of desired datetime
    :type month: int
    :param month: month of desired datetime
    :type day: int
    :param day: day of desired datetime
    :type hour: int
    :param hour: hour of desired datetime
    :type minute: int
    :param minute: minute of desired datetime
    :type seconds: int
    :param seconds: seconds of desired datetime
    :type micro_secs: int
    :param micro_secs: micro seconds of desired datetime
    :type hours_from_utc: int
    :param hours_from_utc: hours difference from UTC timezone
    :returns: timestamp in UTC
    """

    dt = datetime.datetime(year, month, day, hour, minute, seconds, micro_secs)
    # getting the timestamp
    ts = datetime.datetime.timestamp(dt)
    # if runs in data bricks - subtract hours_from_utc hours to transfer to relevant time zone
    if is_databricks():
        ts = ts - 3600 * hours_from_utc
    # convert to ms
    ts_in_ms = math.ceil(ts * 1000)
    return ts_in_ms

def mstimestamp_to_timezone(timestamp, timezone='Israel', milli=True, hour=True, return_datetime=False):
    """
    :type timestamp: float / str / int
    :param timestamp: millisecond timestamp
    :type timezone: str
    :param timezone: ZoneInfo Timestamp name, defaults to Israel
    :type milli: bool
    :param milli: if false, omits millisecond from result
    :type return_datetime: bool
    :param return_datetime: whether to return datetime
    :rtype: str | datetime
    :return: String of datetime in timezone | datetime object in timezone
    """
    server_timezone = ZoneInfo("Etc/UTC")
    chosen_timezone = ZoneInfo(timezone)
    if timestamp == 0 or np.isnan(timestamp):
        return None
    try:
        timestamp = float(timestamp)
    except ValueError as e:
        raise ValueError(f'Timestamp {timestamp} could not be converted to float!' + e)
    dt = datetime.datetime.fromtimestamp(timestamp / 1000)
    dt.replace(tzinfo=server_timezone)
    if return_datetime:
        return dt.astimezone(chosen_timezone)
    if hour:
        if milli:
            dt = dt.astimezone(chosen_timezone).strftime('%Y-%m-%d %H:%M:%S.%f %Z')
        else:
            dt = dt.astimezone(chosen_timezone).strftime('%Y-%m-%d %H:%M:%S %Z')
    else:
        dt = dt.astimezone(chosen_timezone).strftime('%Y-%m-%d')
    return dt

def convert_timestamp_to_datetime(timestamp, up_to_sec_res=False):
    """
    converts timestamp to datetime
    :param timestamp: timestamp
    :type timestamp: str
    :param up_to_sec_res: if true will return the datetime in a resolution of seconds
    :type up_to_sec_res: bool
    """
    num_digits = len(str(int(float(timestamp))))
    timestamp = float(timestamp)
    timestamp = timestamp * math.pow(10, ((num_digits*-1) + 10))
    dt = datetime.datetime.fromtimestamp(float(timestamp))
    if up_to_sec_res:
        dt = dt - datetime.timedelta(microseconds=dt.microsecond)
    return dt

def datetime_to_timezone(dt, timezone='Israel'):
    return dt.astimezone(ZoneInfo(timezone))


def current_timestamp():
    """returns current timestamp (UTC) in milliseconds"""
    return datetime.datetime.timestamp(datetime.datetime.now()) * 1000


def timestamp_timedelta(method=False, **kwargs):
    now = datetime.datetime.now()
    if method:
        calc = now + datetime.timedelta(**kwargs)
    else:
        calc = now - datetime.timedelta(**kwargs)
    return calc.timestamp() * 1000


def string_to_bool(string):
    if string == 'True':
        return True
    if string == 'False':
        return False
    else:
        raise ValueError('Value not equal to True or False!')


def filter_namedtuple(namedtuple, keys):
    """
    gets named tuple (from DataFrame.itertuples) and returns a dictionary of filtered keys from named tuple,
    filtering out keys which have NaN values
    :type namedtuple: Named Tuple
    :param namedtuple: named tuple to filter
    :type keys: list
    :param keys: keys to filter from named tuple
    :rtype: dict
    :return: dictionary of keys and values filtered from named tuple
    """
    d = dict()
    for k in keys:
        try:
            val = getattr(namedtuple, k)
            if pd.isna(val):
                continue
            if str(val).replace('.', '').isnumeric():
                val = int(val)
            d[k] = val
        except AttributeError:
            continue
    return d


# External Imports
import pandas as pd
import math
import pytz
from requests import JSONDecodeError
from enum import Enum
from time import sleep
import datetime
import random
import urllib.parse

# Internal Imports
from wiliot_api.api_client import WiliotCloudError
from wiliot_api.platform.platform import PlatformClient
from wiliot_api.edge.edge import EdgeClient, BridgeAction
from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.common.utils_defines import BO_DICT, SEP

EXCEPTIONS_TO_CATCH = (AttributeError, WiliotCloudError, JSONDecodeError)


# Enum Classes
class GatewayType(Enum):
    UNKNOWN = 'unknown'
    WIFI = 'wifi'
    LTE = 'lte'
    MOBILE = 'mobile'
    IOS = 'ios'
    ANDROID = 'android'


class GatewayAction(Enum):
    ADD_SSID = 'addSsid'
    TOGGLE_SSID = 'toggleDefaultSsid'
    REBOOT_GW = 'rebootGw'
    START_BRIDGE_OTA = '!send_msg_to_brg'

class AndroidGatewayAction(Enum):
    DISABLE_UPLINK = -2
    ENABLE_UPLINK = -3
    DISABLE_BLE_LOGS = -4
    ENABLE_BLE_LOGS = -5

class BridgeThroughGatewayAction(Enum):
    REBOOT = '01'
    BLINK = '02'
    POWER_MGMT = '03'
    RESTORE_DEFAULTS = '04'
    SEND_HB = '05'
    PRODUCTION_MODE = '06'
    SPARSE_37 = '07'


class BoardTypes(Enum):
    ENERGOUS = 'Energous Dual-Band (v0)'
    FANSTEL_SINGLE = 'Fanstel Single-Band (v0)'
    FANSTEL_DUAL = 'Fanstel Dual-Band (v0)'
    MINEW_SINGLE = 'Minew Single-Band (v0)'
    MINEW_DUAL = 'Minew Dual-Band (v0)'
    MOKO = 'Moko Dual-Band (v0)'


# API Clients

class ParamMissingError(Exception):
    pass


class ExtendedEdgeClient(EdgeClient):
    # Get Info
    def get_connected_brgs(self, gw, ignore_bridges=None):
        """
            returned all bridges connected to gateway and claimed by owner
            :type gw: string
            :param gw: desired gw to get its connected bridges
            :type ignore_bridges: list
            :param ignore_bridges: list of bridges ID to ignore
            output type: array of dictionaries
            output param: array with a dictionary to every connected bridge.
                            dictionary includes its current configurations
            example:
            print(wiliot.get_connected_brgs("GW0CDC7EDB1674")):
            TODO: add output
        """
        ignore_bridges = ignore_bridges if ignore_bridges is not None else []
        brg_list = self.get_bridges_connected_to_gateway(gw)

        for brg in brg_list:
            if brg["id"] in ignore_bridges:
                brg_list.remove(brg)
        brg_list_connected = [b for b in brg_list if
                              any([c["connected"] and c["gatewayId"] == gw for c in b["connections"]])]
        brg_list_connected = [b for b in brg_list_connected if b['owned']]
        return brg_list_connected

    def get_bridges(self, online=None, gateway_id=None):
        """
        Get all bridges "seen" by gateways owned by the owner
        :param online: A boolean - optional. Allows to filter only online (True) or offline (False) bridges
        :param gateway_id: A string / list - optional. Allows to filter only bridges currently connected to the gateway / gateways
        :return: A list of bridges
        """
        path = "bridge"
        params = {}
        
        # check type
        if gateway_id is not None and type(gateway_id) not in [str, list]:
            raise TypeError('Gateway ID must be either str/list!')
        gws_list = None
        if type(gateway_id) == list:
            gws_list = gateway_id
        if type(gateway_id) == str:
            gws_list = [gateway_id]
        if online is not None and online is not False:
            params['online'] = online
        try:
            res = self._get(path, params=params)
            bridges = res["data"]
            if gws_list is not None:
                bridges = [b for b in bridges if any([c["connected"] and c["gatewayId"] in gws_list for c
                                                      in b["connections"]])]
            else:
                if online is not None and online is not False:
                    bridges = [b for b in bridges if any([c["connected"] for c in b["connections"]])]
                else:
                    bridges = [b for b in bridges]
            return bridges
        except WiliotCloudError:
            raise

    def get_gateways_from_bridges(self, brg_ids=None):
        """
        gets connected gateways from bridge IDs
        :type brg_ids: list
        :param brg_ids: bridge IDs
        :rtype: dict
        :return: dict of {gatewayId: [list of bridgeIds]} of all connected bridges / specified brg_ids
        """
        brgs = self.get_bridges(online=False)
        owners_gws = [g['gatewayId'] for g in self.get_gateways()]
        gws_list = {}
        for brg_dict in brgs:
            brg_id = brg_dict['id'] 
            # Filter only bridges in brg_ids
            if brg_ids is not None:
                if brg_dict['id'] not in brg_ids:
                    continue
            for connection in brg_dict['connections']:
                if connection['connected']:
                    gw = connection['gatewayId']
                    if gw not in owners_gws:
                        continue
                    if gw not in gws_list.keys():
                        gws_list[gw] = [brg_id]
                    else:
                        gws_list[gw].append(brg_id)
        return gws_list
            
            
    def get_seen_bridges(self, gw, ignore_bridges=None):
        """
        return all bridges 'seen' by GW
        :type gw: str
        :param gw: gateway ID
        :type ignore_bridges: list
        :param ignore_bridges: list of bridge IDs to ignore
        :rtype: list
        :return: list of Bridge IDs
        """
        bridges = [b['bridgeId'] for b in self.get_gateway(gw)['connections']]
        if ignore_bridges is not None:
            bridges = list(set(bridges) - set(ignore_bridges))
        return bridges

    def get_gateway_type(self, gateway_id):
        """
        Returns GatewayType of gateway ID
        :param gateway_id: String - gateway ID
        :rtype: GatewayType
        """
        type = self.get_gateway(gateway_id)['gatewayType'].upper()
        return GatewayType.__getitem__(type)

    def get_gateways_types(self, gw_ids):
        """
        returns dict of gatewaytype:[gatewayIds]
        :param gw_ids: gateways IDs
        :type gw_ids: list
        """
        res = {}
        for gw in gw_ids:
            try:
                gw_type = self.get_gateway_type(gw)
            except WiliotCloudError: 
                gw_type = GatewayType.UNKNOWN
            if gw_type not in res.keys():
                res[gw_type] = []
            res[gw_type].append(gw)
        return res
    
    def get_owner_gateways_types(self):
        gateways = self.get_gateways()
        gateways_types = {}
        for gw in gateways:
            gw_id = gw['gatewayId']
            gw_type = GatewayType.UNKNOWN
            try:
                gw_type = GatewayType.__getitem__(gw['gatewayType'].upper())
                if gw_type == GatewayType.MOBILE:
                    if '-' in gw_id:
                        gw_type = GatewayType.IOS
                    else:
                        gw_type = GatewayType.ANDROID
            except KeyError as e:
                debug_print(f'KeyError {e} when checking GW {gw_id} type')
            gateways_types[gw_id] = gw_type
        return gateways_types

    def check_gw_online(self, gw_ids):
        gws_online = True
        for gw_id in gw_ids:
            gw = self.get_gateway(gw_id)
            if not gw['online']:
                debug_print("Gateway {} is offline".format(gw_id))
                gws_online = False
            else:
                debug_print(f'Gateway {gw_id} is online.')
        return gws_online

    def test_if_bridge_connected(self, brg):
        """
            :type brg: dictionary
            :param brg: bridge data structure
            :return: True if bridge is connected, False otherwise
        """
        for connection in brg['connections']:
            if connection['connected']:
                return True
        return False

    def get_bridge_board(self, brg_id):
        """
        :type brg_id: string
        :param brg_id: bridge id
        """
        return self.get_bridge(brg_id)['boardType']

    # Gateway & Bridge Actions
    def send_action_to_gateway(self, gateway_id, action, **kwargs):
        """
        Send an action to a gateway
        :param gateway_id: String - the ID of the gateway to send the action to
        :param action: GatewayAction - Required
        :return: True if the cloud successfully sent the action to the gateway, False otherwise
        """
        gw_type = self.get_gateway_type(gateway_id)
        assert gw_type in [GatewayType.WIFI, GatewayType.MOBILE], 'gateway does not support action API!'
        if gw_type == GatewayType.WIFI:
            assert action in GatewayAction, 'action not valid for WIFI GW!'
        if gw_type == GatewayType.MOBILE:
            assert action in AndroidGatewayAction, 'action not valid for Mobile GW!'
        if action == GatewayAction.ADD_SSID:
            assert 'ssid' in kwargs.keys(), "Missing a 'ssid' parameter"
        if action == GatewayAction.START_BRIDGE_OTA:
            assert 'bridge_id' in kwargs.keys(), "Missing a 'bridge_id' parameter"
        path = "gateway/{}/action".format(gateway_id)
        action_payload = action.value
        if action == GatewayAction.START_BRIDGE_OTA:
            action_payload = action_payload + f' {kwargs["bridge_id"]} 1'
        else:
            for key, value in kwargs.items():
                action_payload = action_payload + f' {key}:{value}'
        payload = {
            "action": action_payload
        }
        try:
            res = self._post(path, payload)
            return res['data'].lower().find("ok") != -1
        except WiliotCloudError as e:
            print("Failed to send action to gateway")
            raise WiliotCloudError(
                "Failed to send action to gateway. Received the following error: {}".format(e.args[0]))

    def send_bridge_action_through_gw(self, gateway_id, action_type, payload='0', bridge_id=None, broadcast=False, reps=8, ms_to_send=1200):
        """
        Send an action to a bridge through a gateway
        :param broadcast: whether to broadcast
        :param gateway_id: String - the ID of the gateway to send the action Through
        :param bridge_id: String - the ID of the bridge to send the action to
        :param action_type: BridgeThroughGatewayAction - Required
        :param payload: 28 bytes (hex string)
        :param reps: repetition (wifi gw only)
        :param ms_to_send: milliseconds to send (android gw only)
        :return: True if the cloud successfully sent the action to the gateway, False otherwise
        """
        assert action_type in BridgeThroughGatewayAction, 'Not valid action type'
        assert (bridge_id is not None and broadcast is False) or (bridge_id is None and broadcast is True), \
            'Must supply bridgeId / set broadcast to True'
        assert reps > 0, 'Repetitions must be a positive value!'
        assert ms_to_send > 0, 'Milliseconds to send must be a positive value!'
        assert len(bridge_id) == 12
        payload = str(payload).ljust(28, '0')
        # Packet Header: 0-5 ADVA, 6 Length, 7 AD Type, 8-9 UUID, 10-12 Group ID, 13 MSG type, 14 API Ver.
        header = '1E16C6FC0000ED0707'
        # Randomize Sequence ID
        seq_id = random.getrandbits(8).to_bytes(1, 'big').hex().upper()
        gw_type = self.get_gateway_type(gateway_id)
        if gw_type == GatewayType.MOBILE:
            arg = ms_to_send
        else:
            arg = reps
        payload = {
            "action": f'!sp {header}{seq_id}{bridge_id}{action_type.value}{payload} {arg}'
        }
        path = "gateway/{}/action".format(gateway_id)
        try:
            res = self._post(path, payload)
            return res['data'].lower().find("ok") != -1
        except WiliotCloudError as e:
            print("Failed to send action to gateway")
            raise WiliotCloudError(
                "Failed to send action to gateway. Received the following error: {}".format(e.args[0]))

    def reboot_gateway(self, gateway_id):
        """
        Reboots specified GW
        :param gateway_id: Gateway ID
        :return: True if the cloud successfully sent the action to the gateway, False otherwise
        """
        return self.send_action_to_gateway(gateway_id, GatewayAction.REBOOT_GW)

    def reboot_bridge(self, bridge_id):
        """
        Reboots specified GW
        :param bridge_id: Bridge ID
        :return: True if the cloud successfully sent the action to the gateway, False otherwise
        """
        return self.send_action_to_bridge(bridge_id, BridgeAction.REBOOT)

    # Logs & Acks
    def fetch_logs(self, gateway_id, hrs_back=24):
        """
        fetches gateway logs from cloud
        :param gateway_id: String - the ID of the gateway
        :param hrs_back: Int - How many hours back to query
        :return: List of Dicts of gateway log entries
        """
        assert hrs_back < 144, 'Cannot query for more than 144 hours back!'
        end_timestamp = int(datetime.datetime.now().timestamp())
        start_timestamp = int((datetime.datetime.now() - datetime.timedelta(hours=hrs_back)).timestamp())
        path = f'gateway/{gateway_id}/logs?' + urllib.parse.urlencode(
            {"start": start_timestamp, "end": end_timestamp, 'step': 60})
        try:
            res = self._get(path)
            return res['data']
        except WiliotCloudError as e:
            print("Failed to fetch gateway logs")
            raise WiliotCloudError(
                "Failed to fetch gateway logs. Received the following error: {}".format(e.args[0]))

    def get_gateway_logs(self, gateway_id, hours_back=24):
        """
        fetches gateway logs from gateway
        :type gateway_id: str
        :param gateway_id: Gateway ID
        :type hours_back: int
        :param hours_back: How many hours back to query
        :rtype: Pandas DataFrame
        :return: Gateway Logs
        """

        logs = pd.DataFrame.from_dict(self.fetch_logs(gateway_id, hours_back))
        if len(logs) == 0:
            return None
        logs['type'] = logs['message'].apply(lambda x: x.split(':')[0])
        logs['message'] = logs['message'].apply(lambda x: ''.join(x.split(':')[1:]))
        return logs

    def print_gateway_logs(self, gateway_id):
        """
        prints gateway logs nicely
        :type gateway_id: str
        :param gateway_id: Gateway ID
        """
        logs = self.get_gateway_logs(gateway_id)
        if logs is None:
            print(f'---No logs for GW {gateway_id}---')
            return None
        logs = logs.reindex(index=logs.index[::-1])
        logs = logs.reset_index()
        datetime_format = "%Y-%m-%dT%H:%M:%SZ"

        for row in logs.itertuples():
            timestamp = datetime.datetime.strptime(row.timestamp, datetime_format).replace(tzinfo=pytz.UTC).timestamp()
            local_datetime = datetime.datetime.fromtimestamp(timestamp)
            print(f'---Log No. {row.Index} | {row.level} | {local_datetime}---')
            print(f'*{row.message}')
            print(f'')
    
    def get_gateway_info(self, gateway_id):
        """
        get gateway info from gateway
        :param gateway_id: Gateway ID
        :type gateway_id: str
        :return: gateway info
        :rtype: dict
        """
        res = self.get_gateway(gateway_id)
        return res['gatewayInfo']
                
    def get_acks(self, gateway_id, hours_back=3):
        """
        function gets latest acknowledge packets for each bridge ID (after sending GW action)
        :type gateway_id: str
        :param gateway_id: gateway ID
        :type hours_back: int
        :param hours_back: hours back to query
        :return: last acknowledge packet
        """

        try:
            logs = self.get_gateway_logs(gateway_id, hours_back=hours_back)
        except Exception as e:
            debug_print(f'Exception {e} Caught when getting Acks from GW {gateway_id}')
            logs = None
        if logs is None:
            return None
        gw_type = self.get_gateway_type(gateway_id)
        if gw_type == GatewayType.MOBILE:
            logs = logs[list(map(lambda x: x.startswith(' ReceivedAction'), logs['message']))]
            logs['rawPacket'] = logs['message'].apply(lambda x: x.split('=')[1][16:-1]).str.upper()
        else:
            logs = logs[list(map(lambda x: x.startswith(' RecievedAction') or x.startswith(' ReceivedAction'), logs['message']))]
            logs['rawPacket'] = logs['message'].apply(lambda x: x.split('=')[1][6:-1])
        logs['bridgeId'] = logs['rawPacket'].apply(lambda x: x[:12])
        logs['actionType'] = logs['rawPacket'].apply(lambda x: BridgeThroughGatewayAction(str(x[12:14])))
        logs['payload'] = logs['rawPacket'].apply(lambda x: x[14:])
        return logs

    # Versions & Version-Dependent
    def get_gw_version(self, gw_id):
        """
        :type gw_id: string
        :param gw_id: gateway id
        """
        # TODO - version bug
        cnt = 0
        while cnt <= 30:
            gw = self.get_gateway(gw_id)
            if 'version' in gw['reportedConf']:
                return gw['reportedConf']['version']
            else:
                debug_print('version not in GW reported conf! sleeping for 10 seconds...')
                sleep(10)

    def get_gw_ble_version(self, gw_id):
        """
        :type gw_id: string
        :param gw_id: gateway id
        """
        gw = self.get_gateway(gw_id)
        return gw["reportedConf"]["bleChipSwVersion"]
    
    def get_gw_interface_version(self, gw_id):
        """
        :type gw_id: string
        :param gw_id: gateway id
        """
        gw = self.get_gateway(gw_id)
        return gw["reportedConf"]["interfaceChipSwVersion"]

    def get_brg_ble_version(self, brg_id):
        """
        :type brg_id: string
        :param brg_id: bridge id
        """
        brg = self.get_bridge(brg_id)
        return brg["version"]

    def check_energous_ble_reboot_needed(self, brg_id):
        """
        checks if energous reboot after config change is needed for BLE version
        :rtype: bool
        :return: True if supported, False otherwise
        """
        ble_ver = self.get_brg_ble_version(brg_id).split('.')
        if ((int(ble_ver[0]) > 3) or (int(ble_ver[0]) > 2 and int(ble_ver[1]) > 11) or \
                (int(ble_ver[0]) > 2 and int(ble_ver[1]) > 10 and int(ble_ver[2]) > 38)):
            return False
        else:
            return True

    def get_max_sub1ghzoutputpower(self, brg_id):
        brg = self.get_bridge(brg_id)
        version = brg["version"].split(".")
        if (int(version[0]) > 3) or (int(version[0]) > 2 and int(version[1]) > 7) or \
                (int(version[0]) > 2 and int(version[1]) > 6 and int(version[2]) > 65):
            return 32
        return None

    def is_global_pacing_zone(self, brg_id):
        """
        check if bridge supports global pacing by zone
        :param brg_id: bridge ID
        :type brg_id: str
        :return: True if supports, False if not
        :rtype: bool
        """
        brg = self.get_bridge(brg_id)
        version = brg["version"].split(".")
        if (int(version[0]) > 3) or ((int(version[0]) == 3) and int(version[1]) > 12) or \
                (int(version[0]) == 3 and int(version[1]) == 12 and int(version[2]) > 35):
            return True
        return False

    def get_pacing_param_name(self, gw_id):
        """
        gets name of global pacing / pacing group parameter name in platform
        :rtype: str
        :return: parameter name
        """
        ble_ver = self.get_gw_ble_version(gw_id).split('.')
        if ((int(ble_ver[0]) > 3) or (int(ble_ver[0]) > 2 and int(ble_ver[1]) > 12) or \
                (int(ble_ver[0]) > 2 and int(ble_ver[1]) > 11 and int(ble_ver[2]) > 34)):
            return 'globalPacingGroup'
        else:
            return 'globalPacingEnabled'

    def print_brgs_versions(brg_list, ignore_bridges=None):
        """
        :type brg_list: list of dictionaries
        :param brg_list: bridges to print
        :type ignore_bridges: list
        :param ignore_bridges: list of bridges ID to ignore
        """
        ignore_bridges = ignore_bridges if ignore_bridges is not None else []
        debug_print(SEP)
        debug_print("Bridges versions:")
        for brg in brg_list:
            if brg["id"] in ignore_bridges:
                continue
            debug_print("{} : {}".format(brg["id"], brg["version"]))
        debug_print(SEP)

    def get_tx_period(self, brg_id, rx_tx_period, d_c, is_single=False):
        if is_single:
            return math.ceil(d_c * rx_tx_period + 0.45)
        brg = self.get_bridge(brg_id)
        version = brg["version"].split(".")
        if (int(version[0]) > 3) or (int(version[0]) > 2 and int(version[1]) > 7) or \
                (int(version[0]) > 2 and int(version[1]) > 6 and int(version[2]) > 56):
            return math.ceil(d_c * rx_tx_period)
        return math.ceil(d_c * rx_tx_period + 1.5)

    def is_brg_updated_to_ble_ver(self, brg_id, ble_ver):
        """
        :type brg_id: str
        :param brg_id: bridge ID
        :type ble_ver: str
        :param ble_ver: BLE version
        :rtype: bool
        :return: True if bridge is updated to BLE version, False otherwise
        """
        brg_ble_ver = self.get_brg_ble_version(brg_id)
        if brg_ble_ver == ble_ver:
            return True
        else:
            return False

    def is_brg_updated_to_gw_ver(self, gw_id, brg_id):
        """
        :type gw_id: str
        :param gw_id: gateway id
        :type brg_id: str
        :param brg_id: bridge id
        :rtype: bool
        :returns: true if bridge is updated to gw version, else returns false
        """
        gw_ver = self.get_gw_ble_version(gw_id)
        brg_ver = self.get_brg_ble_version(brg_id)
        if gw_ver == brg_ver:
            debug_print(f'GW {gw_id} and BRG {brg_id} are both updated to version {self.get_gw_version(gw_id)}')
            return True
        else:
            debug_print(f'GW {gw_id} ver. {gw_ver}, BRG {brg_id} ver. {brg_ver}')
            return False

    def check_thin_gw_support(self, gw_id):
        if self.get_gateway_type(gw_id) == GatewayType.WIFI:
            return self.check_ble_thin_gw_support(self.get_gw_ble_version(gw_id))
        else:
            return True

    @staticmethod
    def check_ble_thin_gw_support(ble_ver):
        """
        checks if thin GW is supported for BLE version
        :rtype: bool
        :return: True if supported, False otherwise
        """
        ble_ver = ble_ver.split('.')
        if not ((int(ble_ver[0]) > 3) or
                (int(ble_ver[0]) > 2 and int(ble_ver[1]) > 11) or \
                (int(ble_ver[0]) > 2 and int(ble_ver[1]) > 10 and int(ble_ver[2]) > 39)):
            return False
        else:
            return True

    @staticmethod
    def ep_36_exist(version):
        """
        :type version: string
        :param version: FW version of bridge
        """
        major, minor, build = [int(v) for v in version]
        if major > 3 or major == 3 and minor > 6 or major == 3 and minor == 6 and build > 28:
            return True
        else:
            return False

    # Configuration Changes
    def change_gw_config(self, gws_list, config_dict, minutes_timeout=5, ignore_missing_params=False):
        """
            change configuration for multiple GWs
            :type gws_list: every iterable type
            :param gws_list: desired gateways to configure
            :type config_dict: dict
            :param config_dict: dictionary of parameters and values to configure
            :type ignore_missing_params: bool
            :param ignore_missing_params: if True, will return all GWs as updated as long as all available params have been changed
        """
        if not config_dict:  # check there are parameters to configure
            return gws_list

        def check_updated(gw_id, config_dict):
            """
            :param gw_id: Gateway ID
            :param config_dict: config dictionary
            """
            gw = self.get_gateway(gw_id)
            if 'version' in config_dict.keys():
                try:
                    gw_version = gw['version']
                    if gw_version != config_dict['version']:
                        return False
                except KeyError:
                    debug_print(f'Version not yet updated for {gw_id}... sleeping for 30 seconds')
                    sleep(30)
            if 'additional' in gw['reportedConf'].keys():
                gw_dict = gw['reportedConf']['additional']
            else:
                gw_dict = {}
            relevant_dict = {}
            missing_keys = []
            for key in config_dict.keys():
                if key not in gw_dict.keys():
                    missing_keys.append(key)
                else:
                    relevant_dict.update({key: gw_dict[key]})
            if relevant_dict == config_dict:
                return True
            if len(missing_keys) > 0:
                if ignore_missing_params:
                    relevant_config_dict = config_dict.copy()
                    for key in missing_keys:
                        relevant_config_dict.pop(key)
                    if relevant_dict == relevant_config_dict:
                        return True
                    return False
                raise ParamMissingError(
                    f"{missing_keys} not in GW {gw_id} Parameters! Check FW Version or change desired config")
            return False

        desired_version = None
        if 'version' in config_dict.keys():
            desired_version = config_dict.pop('version')
        config_dict_altered = {'additional': config_dict}
        if desired_version is not None:
            config_dict_altered.update({'version': desired_version})
        updated_gws = []
        params_missing = []
        config_sent = False
        timeout = datetime.datetime.now() + datetime.timedelta(minutes=minutes_timeout)
        sleep_needed = False
        while datetime.datetime.now() <= timeout and set(updated_gws) != set(gws_list):
            if sleep_needed:
                sleep(10)
                sleep_needed = False
            # Configure Gateways
            if not config_sent:
                try:
                    self.update_gateways_configuration(gws_list, config_dict_altered)
                    debug_print(f'Configuration sent to {gws_list}')
                    config_sent = True
                except EXCEPTIONS_TO_CATCH as e:
                    debug_print(f'{gws_list}: {e}')
                    sleep_needed = True

            # Check if GWs updated
            for gw_id in set(gws_list) - set(updated_gws):
                try:
                    gw_updated = check_updated(gw_id, config_dict)
                    if gw_updated:
                        updated_gws.append(gw_id)
                except EXCEPTIONS_TO_CATCH as e:
                    debug_print(f'{gw_id}: {e}')
                    sleep_needed = True
                except ParamMissingError as e:
                    params_missing.append((e, gw_id))
                    updated_gws.append(gw_id)
        for error, gw in params_missing:
            debug_print(error)
            updated_gws.remove(gw)
        debug_print(f'{updated_gws} updated to desired configuration:')
        debug_print(config_dict, pretty=True)
        return updated_gws

    def change_brg_config(self, connected_bridges, config_dict, minutes_timeout=5, ignore_bridges=[None]):
        """
            change configuration for all bridge in connected_bridges
            hange configuration to all in connected_bridges
            :type connected_bridges: every iterable type
            :param connected_bridges: desired bridges to configure
            :type config_dict: dict
            :param config_dict: dictionary of parameters and values to configure
            :type ignore_bridges: list
            :param ignore_bridges: list of bridges ID to ignore
            :rtype: list
            :return: list of updated bridges
        """
        brgs_list = list(set(connected_bridges) - set(ignore_bridges))
        updated_brgs = []
        if len(brgs_list) == 0:
            return updated_brgs
        if not config_dict:
            return brgs_list

        def update_bridge(brg_id, config_dict):
            """
            :param brg_id: Bridge ID
            :param config_dict: config dictionary
            """
            res = self.update_bridge_configuration(brg_id, config_dict)
            if self.get_bridge_board(brg_id) == BoardTypes.ENERGOUS.value and self.check_energous_ble_reboot_needed(
                    brg_id):
                debug_print(f'Rebooting energous Bridge {brg_id}...')
                self.send_action_to_bridge(brg_id, BridgeAction.REBOOT)
            return res

        def check_updated(brg_id, config_dict):
            """
            :param brg_id: Bridge ID
            :param config_dict: config dictionary
            """
            brg = self.get_bridge(brg_id)
            brg_dict = brg['config']
            relevant_dict = {}
            missing_keys = []
            for key in config_dict.keys():
                if key not in brg_dict.keys():
                    missing_keys.append(key)
                else:
                    relevant_dict.update({key: brg_dict[key]})
            if relevant_dict == config_dict:
                debug_print(f'{brg_id} updated')
                return True
            if len(missing_keys) > 0:
                raise ParamMissingError(
                    f"{missing_keys} not in BRG {brg_id} Parameters! Check FW Version or change desired config")
            return False

        timeout = datetime.datetime.now() + datetime.timedelta(minutes=minutes_timeout)
        brgs_to_config = {brg: True for brg in brgs_list}
        params_missing = []
        sleep_needed = False
        while datetime.datetime.now() <= timeout and set(updated_brgs) != set(brgs_list):
            if sleep_needed:
                sleep(10)
                sleep_needed = False
            # Update bridges
            for brg_id in filter(brgs_to_config.get, brgs_to_config):
                try:
                    res = update_bridge(brg_id, config_dict)
                    debug_print(f'Sent configuration to bridge {brg_id}')
                    brgs_to_config[brg_id] = not res
                except EXCEPTIONS_TO_CATCH as e:
                    debug_print(f'{brg_id}: {e}')
                    ueded = True
            # Check if bridges updated
            for brg_id in set(brgs_list) - set(updated_brgs):
                try:
                    brg_updated = check_updated(brg_id, config_dict)
                    if brg_updated:
                        updated_brgs.append(brg_id)
                    else:
                        debug_print(f'Sending configuration again to bridge {brg_id}')
                        sleep(3)
                        update_bridge(brg_id, config_dict)
                except (EXCEPTIONS_TO_CATCH) as e:
                    debug_print(f'{brg_id}: {e}')
                    sleep_needed = True
                except ParamMissingError as e:
                    params_missing.append((e, brg_id))
                    updated_brgs.append(brg_id)
        for error, brg in params_missing:
            debug_print(error)
            updated_brgs.remove(brg)
        if minutes_timeout>0:
            debug_print(f'{(updated_brgs)} updated to desired configuration:')
        else:
            debug_print(f'{brgs_list} will update to desired configuration when back online:')
        debug_print(config_dict, pretty=True)
        return updated_brgs

    # Connect & Brownout
    def change_to_brownout(self, connected_bridges, seconds_in_bo=0, ignore_bridges=None):
        """
            change all bridges in connected_bridges to brownout mode and sleeps for seconds_in_bo seconds.
            :type connected_bridges: every iterable type
            :param connected_bridges: desired bridges to configure
            :type seconds_in_bo: int
            :param seconds_in_bo: time to sleep in brownout [seconds]
            :type ignore_bridges: list
            :param ignore_bridges: list of bridges ID to ignore
            example:
            wiliot.change_to_brownout(["4CEE79FA3523", "D0269CFEB518"], seconds_in_bo=300)
        """
        ignore_bridges = list() if ignore_bridges is None else ignore_bridges
        relevant_brgs = list(set(connected_bridges) - set(ignore_bridges))
        brgs_updated = self.change_brg_config(relevant_brgs, BO_DICT)
        if seconds_in_bo > 0:
            debug_print(f'Sleeping for {datetime.timedelta(seconds=seconds_in_bo)}')
            sleep(seconds_in_bo)
        return brgs_updated

    def connect_gw_bridges(self, gw_ids=None, minutes_timeout=5, expected_num_brgs=None, ignore_bridges=None,
                           do_brown_out=True, expected_brgs=None):
        """
            Iteratively connects to bridges and change them to brownout.
            Function can take either GW IDs as input and then connect to all brgs connected to them 
            and if needed change them to brownout, 
            OR takes expected brgs and then connect to those and if needed change them to brownout.
            brgs listed in ignore bridges will be completely ignored.
            :type gw_ids: list
            :param gw_ids: list of GW ids
            :type minutes_timeout: int
            :param minutes_timeout: timeout to connect to bridges
            :type expected_num_brgs: int
            :param expected_num_brgs: maximum bridges in the deployment.
                                      if None, the function will scan for new bridges until minutes_timeout is over / wait for all expected brgs
            :type ignore_bridges: list
            :param ignore_bridges: list of bridges ID to ignore
            :type do_brown_out: bool
            :param do_brown_out: brown out bridges (change to EP 36)
            :type expected_brgs: list
            :param expected_brgs: list of expected BRG ids
            :rtype board_type_dict: dict
            :return board_type_dict: dictionary of bridge and board type
            :rtype connected_brgs type: list
            :return connected_brgs param: list of all connected bridges from csv
        """
        
        def get_bridge_type(brg_dict):
            if 'single' in brg_dict['boardType'].lower():
                return 'single'
            return 'dual'

        def get_connected_bridges(gw_ids, ignore_bridges, expected_brgs, timeout):
            connected_bridges = list()
            connected_bridges_dicts = None
            sleep_needed = False
            while connected_bridges_dicts is None and datetime.datetime.now() < timeout:
                if sleep_needed:
                    sleep(5)
                try:
                    connected_bridges_dicts = self.get_bridges(online=True, gateway_id=gw_ids)
                except EXCEPTIONS_TO_CATCH as e:
                    print(e)
                    sleep_needed = True
            board_type_dict = {'dual': [], 'single': []}
            for brg_dict in connected_bridges_dicts:
                brg_id = brg_dict['id']
                # skip ignore bridges
                if brg_id in ignore_bridges:
                    continue
                # skip bridges not in expected brgs
                if len(expected_brgs) > 0:
                    if brg_id not in expected_brgs:
                        continue
                # claim unclaimed bridges
                if not brg_dict['claimed']:
                    self.claim_bridge(brg_id)
                board_type_dict[get_bridge_type(brg_dict)].append(brg_id)
                connected_bridges.append(brg_id)
            return connected_bridges, board_type_dict
        if gw_ids is None:
            if len(expected_brgs) > 0:
                gw_ids = list(self.get_gateways_from_bridges(expected_brgs).keys())
            else:
                raise ValueError('Must input either GW IDs or expected bridges!')
        ignore_bridges = list() if ignore_bridges is None else ignore_bridges
        expected_brgs = list() if expected_brgs is None else expected_brgs
        # remove ignore bridges from expected_brgs
        expected_brgs = list(set(expected_brgs) - set(ignore_bridges))
        expected_num_brgs = -1 if expected_num_brgs is None else expected_num_brgs
        timeout = datetime.datetime.now() + datetime.timedelta(minutes=minutes_timeout)
        connected_bridges, board_type_dict = get_connected_bridges(
            gw_ids, ignore_bridges, expected_brgs, timeout)
        brgs_brownout_status = dict()
        if do_brown_out:
            brgs_brownout_status = dict(map(lambda x: (x, False), connected_bridges))
        if expected_num_brgs < 0:
            debug_print(
                f'Connecting bridges | Timeout at {timeout.time()}')
        else:
            debug_print(
                f'Connecting{" and browning out" if do_brown_out else ""} {expected_num_brgs} bridges | Timeout at {timeout.time()}')
        if len(expected_brgs) > 0:
            debug_print(f'Expecting bridges {expected_brgs}')
            if expected_num_brgs == -1:
                expected_num_brgs = len(expected_brgs)
        last_print = datetime.datetime.now() - datetime.timedelta(minutes=1)
        while (datetime.datetime.now() < timeout and \
               (False in brgs_brownout_status.values() or (len(connected_bridges) < expected_num_brgs
                                                           and not (set(expected_brgs) <= set(connected_bridges))))):
            connected_bridges, board_type_dict = get_connected_bridges(
                gw_ids, ignore_bridges, expected_brgs, timeout)
            # update new brgs to brownout
            new_brgs = list(set(connected_bridges) - set(brgs_brownout_status.keys()))
            brgs_brownout_status.update({brg: False for brg in new_brgs})
            if (datetime.datetime.now() - last_print).total_seconds() > 5:
                debug_print(
                    f'{len(connected_bridges)}{(" / " + str(expected_num_brgs)) if expected_num_brgs > 0 else ""} connected bridges')
                last_print = datetime.datetime.now()
            # do brown out for needed bridges
            if do_brown_out and False in brgs_brownout_status.values():
                # update only needed bridges
                brgs_to_update = [brg for brg, status in brgs_brownout_status.items() if status is False]
                brgs_updated = self.change_to_brownout(brgs_to_update)
                # change brownout status to true
                brgs_brownout_status.update({brg: True for brg in brgs_updated})
        if False in brgs_brownout_status.values():
            browned_out_brgs = list(filter(brgs_brownout_status.get, brgs_brownout_status))
            non_browned_out_brgs = list(set(connected_bridges) - set(browned_out_brgs))
            raise WiliotCloudError(f'Cannot change connected BRGs {non_browned_out_brgs} to brownout! Check deployment!')
        
        debug_print(f'{len(gw_ids)} GWs {gw_ids} Connected')
        debug_print(f'{len(connected_bridges)} BRGs {connected_bridges} Connected')
        return connected_bridges, board_type_dict
    
    
    def check_gw_compatible_for_action(self, gateway_id):
        """
        checks if gateway is compatible to send and receive power management configurations
        :type gateway_id: str
        :param gateway_id: gateway ID
        :rtype: bool
        """
        gateway_type = self.get_gateway_type(gateway_id)
        if gateway_type != GatewayType.WIFI:
            if gateway_type == GatewayType.MOBILE: 
                return True
            else:
                return False
        gw_dict = self.get_gateway(gateway_id)        
        ble_ver = gw_dict["reportedConf"]["bleChipSwVersion"]
        if gw_dict['online']:
            if gw_dict['reportedConf']['additional']['gwMgmtMode'] == 'transparent':
                support = self.check_gw_ble_transparent_support(ble_ver)
                if not support:
                    debug_print(f'Please update GW {gateway_id} or change to active mode for power management support!')
                return support
            return True
        return False
    
    @staticmethod
    def check_gw_ble_transparent_support(ble_ver):
        """
        checks if power management is supported for transparent GW BLE version
        :rtype: bool
        :return: True if supported, False otherwise
        """
        ble_ver = ble_ver.split('.')
        if not ((int(ble_ver[0]) > 3) or
                (int(ble_ver[0]) > 2 and int(ble_ver[1]) > 14) or \
                (int(ble_ver[0]) > 2 and int(ble_ver[1]) > 13 and int(ble_ver[2]) > 27)):
            return False
        else:
            return True



class ExtendedPlatformClient(PlatformClient):
    def __init__(self, api_key, owner_id, env='', log_file=None, logger_=None):
        super().__init__(api_key, owner_id, env, log_file, logger_)

    def get_locations(self, locations=None):
        """
        :rtype: dict
        :return: Dictionary of locations and zones with relevant associations for each
        """
        in_locations = locations
        locations = super().get_locations()
        res = []
        for location in locations:
            if in_locations is not None and location['name'] not in in_locations:
                continue
            loc_id = location['id']
            location['associations'] = super().get_location_associations(loc_id)
            zones = super().get_zones(loc_id)
            for zone in zones:
                zone_id = zone['id']
                zone['associations'] = super().get_zone_associations(zone_id)
            location['zones'] = zones
            res.append(location)
        return res

    def get_locations_bridges(self, locations=None):
        """
            change a list of bridges associated to location
            if location is None, reffer all owner's locations
            :type locations: every iterable type
            :param locations: desired locations
            returns a list of dictionaries
            Each dictionary includes bridgeId, location properties and zone properties if a bridge is associated to zone
        """
        locations = self.get_locations(locations=locations)
        bridges = []
        for location in locations:
            # Add bridges directly associated with location
            for bridge in location['associations']:
                if bridge['associationType'] == 'bridge':
                    bridges.append({'bridgeId': bridge['associationValue'],
                                    'locationId': location['id'],
                                    'locationName': location['name'],
                                    'locationType': location['locationType']})
                    if 'location' in location.keys():
                        bridges[-1].update({
                            'locationLat': location['lat'],
                            'locationLng': location['lng'],
                            'location': location['location'],
                            'locationAddress': location['address']})
            # add bridges associated to zone (inside location)
            for zone in location['zones']:
                for bridge in zone['associations']:
                    if bridge['associationType'] == 'bridge':
                        bridges.append({'bridgeId': bridge['associationValue'],
                                        'locationId': location['id'],
                                        'locationName': location['name'],
                                        'zoneId': zone['id'],
                                        'zoneName': zone['name'],
                                        'locationType': location['locationType']})
                        if 'location' in location.keys():
                            bridges[-1].update({
                                'locationLat': location['lat'],
                                'locationLng': location['lng'],
                                'location': location['location'],
                                'locationAddress': location['address']})
        return bridges

    def get_location_id(self, location_name):
        locations = self.get_locations()
        for loc_dict in locations:
            if loc_dict['name'] == location_name:
                return loc_dict['id']

    def get_location_bridges(self, location_name):
        """
        get brgs from location name
        :param location_name: name of location
        :type location_name: str
        :return: list of brgs in location
        :rtype: list
        """
        location_id = self.get_location_id(location_name)
        loc_associations = self.get_location_associations(location_id)
        brgs_list = [assoc['associationValue'] for assoc in loc_associations if (assoc['associationType'] == 'bridge')]
        return brgs_list

    def get_locations_names(self):
        locations = super().get_locations()
        names = []
        for loc in locations:
            names.append(loc['name'])
        return names


import datetime
import logging
import os
import colorama
import tabulate
import pprint
from git import Repo, InvalidGitRepositoryError

import wiliot_deployment_tools
from wiliot_deployment_tools.utils.get_version import get_version


# Debugging & Printing
def print_package_gitinfo():
    """prints the current branch, latest commit datetime and commit hash"""
    version = get_version()
    try:
        repo = Repo(wiliot_deployment_tools.__file__[:-35])
        commit_hash = repo.git.rev_parse("HEAD")
        branch = repo.active_branch
        date = repo.head.object.committed_datetime
        debug_print(f'Wiliot Deployment Tools branch {branch} | latest commit at {date} | {commit_hash}')
        return True
    except InvalidGitRepositoryError:
        debug_print(f"Wiliot Deployment Tools version {version}")
        return True

def debug_print(txt, pretty=False, center=False, tab=False, color=None):
    """
    :type txt: string
    :param txt: text to print
    :type pretty: bool
    :param pretty: pretty print
    :type center: bool
    :param center: print with stars (*)
    :type tab: bool
    :param tab: tabulate input
    :type color: str
    :param color: colorama color code
    """
    if tab:
        txt = '\n' + tabulate.tabulate(txt, tablefmt='rst')
    if type(txt) != str:
        try:
            txt = str(txt)
        except TypeError:
            debug_print('Could not Print!')
    if center:
        txt = txt.center(94, '-')
    if color is not None:
        txt = eval(f'colorama.Fore.{color}') + txt + colorama.Style.RESET_ALL
    if pretty:
        txt = pprint.pformat(txt, sort_dicts=False)

    # Print Text
    if logging.getLogger().hasHandlers():
        log = logging.getLogger()
        log.info(txt)
    else:
        print(datetime.datetime.now().strftime("[%d/%m/%Y, %H:%M:%S]: ") + txt)

def is_databricks():
    """returns if running in databricks"""
    try:
        get_ipython().__class__.__name__
        return True
    except NameError:
        return False


def is_databricks():
    try:
        from IPython import get_ipython
        if "IPKernelApp" not in get_ipython().config:
            return False
        if "VSCODE_PID" in os.environ:
            return False
        return True
    except Exception as e:
        return False



from wiliot_deployment_tools.calibration_mgmt.calibration_mgmt import CalibrationManagementClient
from wiliot_core import check_user_config_is_ok
from wiliot_deployment_tools.common.utils_defines import *
from argparse import ArgumentParser
import sys
import bitstruct


def main():
    # parser
    parser = ArgumentParser(prog='wlt-calibration-mgmt',
                            usage='%(prog)s -o OwnerID -brg BridgeID (-t) [mode]',
                            description='CLI Tool for using Wiliot Bridge calibration management functionality')
    parser.add_argument('-owner', type=str, help="Owner ID", required=True)
    parser.add_argument('-brg', type=str, help="Bridge ID", required=False)
    parser.add_argument('-t', action='store_true',
                        help='if flag used, use test environment (prod is used by default)')
    parser.add_argument('-gw', type=str, help="GW ID to configure bridge (required only for broadcast mode)", required=False)
    parser.add_argument('-m', type=int, help="Minutes timeout (not required, defaults to 5 minutes)", required=False, default=5)
    parser.add_argument('-mode', '--option', choices=[0,1,2], type=int, help="Calinration mode selection: 0 - regular 1 - 38,38,39 2 - 37 on data only", required=True)

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args, unknown = parser.parse_known_args()
    if args.t:
        env = 'test'
    else:
        env = 'prod'
    owner_id = args.owner
    conf_env = env if env == 'prod' else 'non-prod'
    user_config_file_path, api_key, is_success = check_user_config_is_ok(owner_id, conf_env, 'edge')


    user_config_file_path, api_key, is_success = check_user_config_is_ok(owner_id, env, 'edge')
    if is_success:
        print('credentials saved/upload from {}'.format(user_config_file_path))
    else:
        raise Exception('invalid credentials - please try again to login')

    client = CalibrationManagementClient(api_key, owner_id, env)
    client.send_calibration_packet_until_ack(gateway_id = args.gw, bridge_id = args.brg, minutes_timeout=args.m,mode = args.option)


def main_cli():
    main()
if __name__ == '__main__':
    main()



from wiliot_deployment_tools.common.utils import *
import time
import datetime

IS_DATABRICKS = is_databricks()


class ConfigParamsException(Exception):
    def __init__(self, value_name, value):
        """
        :type value_name: string
        :param value_name: value name as appears fo the user
        :type value: string
        :param value: value received
        """
        raise Exception("Please enter valid value to the field '" + str(value_name) +
                        "' (received value is " + str(value) + ")")


class ConfigureDevices:
    if IS_DATABRICKS:
        db_utils().widgets.removeAll()
        db_utils().widgets.text('Edge Management API Key', '')
        db_utils().widgets.dropdown('Environment', 'prod', ['prod', 'test'])
        db_utils().widgets.dropdown('Client', 'Platform', ['Platform', 'Management'])
        db_utils().widgets.text("Owner", '')

        db_utils().widgets.dropdown("Enter or Exit contextual energizing", 'Enter', ['Enter', 'Exit'])
        db_utils().widgets.text("Bridges to configure", '')

        db_utils().widgets.text("Energizing time (in minutes)", '')
        db_utils().widgets.text("Idle time (in minutes)", '')
        db_utils().widgets.text("Start time for energizing (hour, minutes)", '8, 0')

    else:
        debug_print('Running locally, make sure to pass all params for test to run successfully')

    def __init__(self):
        """"
        initialize and save parameters from user
        """
        def check_legal_values():
            if self.enter_or_exit not in ['Enter', 'Exit']:
                raise ConfigParamsException("Enter or Exit contextual energizing ('Enter', 'Exit')",
                                            str(self.enter_or_exit))
            if self.brgs_ids is None:
                raise ConfigParamsException("Bridges to configure", str(self.brgs_ids))

            if self.start_time is None:
                raise ConfigParamsException("Start time for energizing (hour, minutes)", str(self.start_time))
            if len(self.start_time) is not 2:
                raise Exception("Not enough values in 'Start time for energizing (hour, minutes)'")

        if IS_DATABRICKS:
            api_key = db_utils().widgets.get("Edge Management API Key")
            env = db_utils().widgets.get("Environment")
            owner = db_utils().widgets.get('Owner')
            self.enter_or_exit = db_utils().widgets.get('Enter or Exit contextual energizing')
            self.on_duration = db_utils().widgets.get('Energizing time (in minutes)')
            self.idle_duration = db_utils().widgets.get('Idle time (in minutes)')

            self.brgs_ids = db_utils().widgets.get("Bridges to configure").replace(' ', '')
            self.brgs_ids = "".join(self.brgs_ids.split())
            self.brgs_ids = None if self.brgs_ids == '' else self.brgs_ids.split(",")

            self.start_time = db_utils().widgets.get("Start time for energizing (hour, minutes)").replace(' ', '')
            self.start_time = "".join(self.start_time.split())
            self.start_time = None if self.start_time == '' else self.start_time.split(",")

        check_legal_values()
        self.contextual_energizing_client = ContextualEnergizingClient(api_key, owner, env)
        self.sleep_until(int(self.start_time[0]), int(self.start_time[1]))
        # TODO - add the configuration part here, using self.contextual_energizing_client.exit_contextual_energizing()

    def sleep_until(self, hour, minute):
        """
        will wait until the desired time inserted to function (supports up to 24 hours ahead)
        :type hour: int
        :param hour: hour to start working (24-hour scale)
        :type minute: int
        :param minute: minute to start working (60-minute scale)
        """
        run_start_time = datetime.datetime.today()
        future = datetime.datetime(run_start_time.year, run_start_time.month, run_start_time.day, hour, minute)
        if future < run_start_time:
            try:
                future = \
                    datetime.datetime(run_start_time.year, run_start_time.month, run_start_time.day + 1, hour, minute)
            except Exception():
                try:
                    future = datetime.datetime(run_start_time.year, run_start_time.month + 1, 1, hour, minute)
                except Exception():
                    future = datetime.datetime(run_start_time.year + 1, 1, 1, hour, minute)

        debug_print("Waiting to the desired time to start configuration process")
        time.sleep(min((future - run_start_time).total_seconds(), 24 * 60 * 60))


if __name__ == '__main__':
    pass



import logging
import os
import shutil
import subprocess
import datetime
from itertools import combinations, chain
from time import sleep
from pathlib import Path
import inspect
from appdirs import user_data_dir

import plotly.express as px
import pandas as pd
import numpy as np
from plotly.offline import plot
from wiliot_deployment_tools.common.debug import debug_print, is_databricks
from os.path import exists
import tabulate
try:
    from wiliot_deployment_tools.internal.utils.network_packet_list import NetworkPacketList, network_data_preparation
    print('NetworkPacketList imported (running internal version)')
except ImportError:
    pass
from wiliot_deployment_tools.common.utils import convert_timestamp_to_datetime, current_timestamp, mstimestamp_to_timezone


if is_databricks():
    logging.getLogger("py4j.java_gateway").setLevel(logging.ERROR)

class WiliotTableError(Exception):
    pass

class WiliotDatabricksUtils:

    def __init__(self, spark):
        self.spark = spark
        self.is_databricks = is_databricks()

    def get_seen_tags(self, table, start_time, end_time, tags_list=None, bridges_list=None, gateways_list=None):
        """
        does an SQL query of the packet data table between specified timestamps
        filters the data for specified tags/bridges/gateways (if specified)
        returns relevant values from the table
        :type table: str
        :param table: name of data table
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :type tags_list: list
        :param tags_list: list of tags to filter from the data
        :type bridges_list: list
        :param bridges_list: list of bridges to filter from the data
        :type gateways_list: list
        :param gateways_list: list of gateways to filter from the data
        :rtype: pandas DataFrame
        :return: dataframe of data from table
        """
        # TODO - debug the SQL Query with different types of lists
        if bridges_list is not None:
            if len(bridges_list) == 1:
                bridges_list = list(bridges_list) + [""]
            bridges_list = tuple(bridges_list)
        if gateways_list is not None:
            if len(gateways_list) == 1:
                gateways_list = list(gateways_list) + [""]
            gateways_list = tuple(gateways_list)
        if tags_list is not None:
            if len(tags_list) == 1:
                tags_list = list(tags_list) + [""]
            tags_list = tuple(tags_list)
        # adding search by date to improve search time (most tables in data bricks partition is by date)
        query_start_datetime = convert_timestamp_to_datetime(str(start_time))
        query_end_datetime = convert_timestamp_to_datetime(str(end_time))
        start_datetime = query_start_datetime - datetime.timedelta(hours=24)
        end_datetime = query_end_datetime + datetime.timedelta(hours=24)
        start_date = datetime.datetime.strftime(start_datetime, '%Y%m%d')
        end_date = datetime.datetime.strftime(end_datetime, '%Y%m%d')
        sql_method = f"""
        select *
        from {table}
        where date between {start_date} and {end_date} 
        and timestamp between {start_time} and {end_time} 
        """
        if tags_list is not None and tags_list != ():
            sql_method = sql_method + f"""and externalId in {tags_list}
        """
        if bridges_list is not None and bridges_list != ():
            sql_method = sql_method + f"""and bridgeId in {bridges_list}
        """
        if gateways_list is not None and gateways_list != ():
            sql_method = sql_method + f"""and gatewayId in {gateways_list}
        """
        if self.spark is not None:
            debug_print('Running SQL query...', center=True)
            debug_print(sql_method)
            query_data = self.spark.sql(sql_method)
            query_data = query_data.toPandas()
            return query_data
        else:
            raise EnvironmentError("SQL query can only run in databricks")

    def get_seen_events(self, table, start_time, end_time, bridges_list=None, gateways_list=None, platform=False):
        """
        does an SQL query of the packet data table between specified timestamps
        filters the data for specified tags/bridges/gateways (if specified)
        returns relevant values from the table
        :type table: str
        :param table: name of data table
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :type bridges_list: list
        :param bridges_list: list of bridges to filter the data by
        :type gateways_list: list
        :param gateways_list: list of gateways to filter the data by
        :type platform: bool
        :param platform: platform/management (true/false)
        :rtype: pandas DataFrame
        :return: dataframe of data from table
        """
        # TODO - debug the SQL Query with different types of lists
        devices_list = ()
        if bridges_list is not None:
            if len(bridges_list) == 1:
                devices_list = list(bridges_list) + [""]
            devices_list = tuple(devices_list)
        if gateways_list is not None:
            if len(gateways_list) == 1 and devices_list == ():
                devices_list = list(gateways_list) + [""]
            elif devices_list is not None:
                devices_list = devices_list + tuple(gateways_list)
            devices_list = tuple(devices_list)

        # adding search by date to improve search time (most tables in data bricks partition is by date)
        query_start_datetime = convert_timestamp_to_datetime(str(start_time))
        query_end_datetime = convert_timestamp_to_datetime(str(end_time))
        start_datetime = query_start_datetime - datetime.timedelta(hours=36)
        end_datetime = query_end_datetime + datetime.timedelta(hours=36)
        start_date = datetime.datetime.strftime(start_datetime, '%Y%m%d')
        end_date = datetime.datetime.strftime(end_datetime, '%Y%m%d')
        sql_method = f"""
        select *
        from {table}
        where date between {start_date} and {end_date} 
        and start between {start_time} and {end_time} 
        """
        if not platform:
            connectivity_event = "name = 'NTWK'"
            id_filter = "id"
        else:
            connectivity_event = "eventName = 'connectivity'"
            id_filter = "assetId"

        if devices_list is not None:
            sql_method = sql_method + f"""and ({connectivity_event} or {id_filter} in {devices_list})
        """
        else:
            sql_method = sql_method + f"""and {connectivity_event}"""
        if self.spark is not None:
            debug_print('Running SQL query...', center=True)
            debug_print(sql_method)
            query_data = self.spark.sql(sql_method)
            query_data = query_data.toPandas()
            return query_data
        else:
            raise EnvironmentError("SQL query can only run in databricks")

    def get_sequence_id_data(self, table, start_time, end_time, gateways_list=None):
        """
        does an SQL query of the packet data table between specified timestamps
        returns only sequence id
        :type table: str
        :param table: name of data table
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :type gateways_list: list
        :param gateways_list: list of gateways to filter the data by
        :rtype: pandas DataFrame
        :return: dataframe of data from table
        """

        if gateways_list is not None:
            if len(gateways_list) == 1:
                gateways_list = list(gateways_list) + [""]
            gateways_list = tuple(gateways_list)
        # adding search by date to improve search time (most tables in data bricks partition is by date)
        query_start_datetime = convert_timestamp_to_datetime(str(start_time))
        query_end_datetime = convert_timestamp_to_datetime(str(end_time))
        start_datetime = query_start_datetime - datetime.timedelta(hours=36)
        end_datetime = query_end_datetime + datetime.timedelta(hours=36)
        start_date = datetime.datetime.strftime(start_datetime, '%Y%m%d')
        end_date = datetime.datetime.strftime(end_datetime, '%Y%m%d')
        sql_method = f"""
        select sequenceId, timestamp, gatewayId, gatewayName
        from {table}
        where date between {start_date} and {end_date} 
        and timestamp between {start_time} and {end_time} 
        """
        if gateways_list is not None:
            sql_method = sql_method + f"""and gatewayId in {gateways_list}
        """
        if self.spark is not None:
            debug_print('Running SQL query...', center=True)
            debug_print(sql_method)
            query_data = self.spark.sql(sql_method)
            query_data = query_data.toPandas()
            return query_data
        else:
            raise EnvironmentError("SQL query can only run in databricks")

    def get_statistics_data(self, table, start_time, end_time, gateways_list=None):
        """
        does an SQL query of the statistics data table between specified timestamps
        :type table: str
        :param table: name of data table
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :type gateways_list: list
        :param gateways_list: list of gateways to filter the data by
        :rtype: pandas DataFrame
        :return: dataframe of data from table
        """

        if gateways_list is not None:
            if len(gateways_list) == 1:
                gateways_list = list(gateways_list) + [""]
            gateways_list = tuple(gateways_list)
        # adding search by date to improve search time (most tables in data bricks partition is by date)
        query_start_datetime = convert_timestamp_to_datetime(str(start_time))
        query_end_datetime = convert_timestamp_to_datetime(str(end_time))
        start_datetime = query_start_datetime - datetime.timedelta(hours=36)
        end_datetime = query_end_datetime + datetime.timedelta(hours=36)
        start_date = datetime.datetime.strftime(start_datetime, '%Y%m%d')
        end_date = datetime.datetime.strftime(end_datetime, '%Y%m%d')
        sql_method = f"""
        select *
        from {table}
        where date between {start_date} and {end_date} 
        and timestamp between {start_time} and {end_time} 
        """
        if gateways_list is not None:
            sql_method = sql_method + f"""and gatewayId in {gateways_list}
        """
        if self.spark is not None:
            debug_print('Running SQL query...', center=True)
            debug_print(sql_method)
            query_data = self.spark.sql(sql_method)
            query_data = query_data.toPandas()
            return query_data
        else:
            raise EnvironmentError("SQL query can only run in databricks")

    def get_heartbeat_data(self, table, start_time, end_time, gateways_list=None, bridges_list=None):
        """
        does an SQL query of the statistics data table between specified timestamps
        :type table: str
        :param table: name of data table
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :type gateways_list: list
        :param gateways_list: list of gateways to filter the data by
        :rtype: pandas DataFrame
        :return: dataframe of data from table
        """

        if gateways_list is not None:
            if len(gateways_list) == 1:
                gateways_list = list(gateways_list) + [""]
            gateways_list = tuple(gateways_list)
        
        if bridges_list is not None:
            if len(bridges_list) == 1:
                bridges_list = list(bridges_list) + [""]
            bridges_list = tuple(bridges_list)
        
        # adding search by date to improve search time (most tables in data bricks partition is by date)
        query_start_datetime = convert_timestamp_to_datetime(str(start_time))
        query_end_datetime = convert_timestamp_to_datetime(str(end_time))
        start_datetime = query_start_datetime - datetime.timedelta(hours=36)
        end_datetime = query_end_datetime + datetime.timedelta(hours=36)
        start_date = datetime.datetime.strftime(start_datetime, '%Y%m%d')
        end_date = datetime.datetime.strftime(end_datetime, '%Y%m%d')
        sql_method = f"""
        select *
        from {table}
        where date between {start_date} and {end_date} 
        and timestamp between {start_time} and {end_time} 
        """
        if gateways_list is not None:
            sql_method = sql_method + f"and gatewayId in {gateways_list}\n"
        if bridges_list is not None:
            sql_method = sql_method + f"and bridgeId in {bridges_list}\n"
        if self.spark is not None:
            debug_print('Running SQL query...', center=True)
            debug_print(sql_method)
            query_data = self.spark.sql(sql_method)
            query_data = query_data.toPandas()
            return query_data
        else:
            raise EnvironmentError("SQL query can only run in databricks")

    
    def get_num_seen_bridges(self, table, start_time, end_time, gateways_list=None):
        """
        gets number of unique bridges seen by each gateways (bridge has to send data packets)
        :type table: str
        :param table: name of data table
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :type gateways_list: list
        :param gateways_list: list of gateways to filter the data by
        :return: dictionary of number of seen bridges per gateway
        """
        if gateways_list is not None:
            if len(gateways_list) == 1:
                gateways_list = list(gateways_list) + [""]
            gateways_list = tuple(gateways_list)
        # adding search by date to improve search time (most tables in data bricks partition is by date)
        query_start_datetime = convert_timestamp_to_datetime(str(start_time))
        query_end_datetime = convert_timestamp_to_datetime(str(end_time))
        start_datetime = query_start_datetime - datetime.timedelta(hours=36)
        end_datetime = query_end_datetime + datetime.timedelta(hours=36)
        start_date = datetime.datetime.strftime(start_datetime, '%Y%m%d')
        end_date = datetime.datetime.strftime(end_datetime, '%Y%m%d')
        sql_method = f"""
        select gatewayId, count(distinct bridgeId) as countBridge
        from {table}
        where date between {start_date} and {end_date} 
        and timestamp between {start_time} and {end_time} 
        """
        if gateways_list is not None:
            sql_method = sql_method + f"""and gatewayId in {gateways_list}
        """
        sql_method = sql_method + "group by gatewayId"
        if self.spark is not None:
            debug_print('Running SQL query...', center=True)
            debug_print(sql_method)
            query_data = self.spark.sql(sql_method)
            query_data = query_data.toPandas()
            query_data = dict(zip(query_data['gatewayId'], query_data['countBridge']))
            return query_data
        else:
            raise EnvironmentError("SQL query can only run in databricks")


    def get_sequence_id_loss(self, packets_table, statistics_table, start_time, end_time, gateways_list=None, sync_tables=True):
        """
        calculates sequence ID loss (takes reboot into consideration) per GW in table
        :type packets_table: str | DataFrame
        :param packets_table: name of data table | data table DataFrame
        :type statistics_table: str | DataFrame
        :param statistics_table: name of statistics table | statistics table DataFrame
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :type gateways_list: list
        :param gateways_list: list of gateways to filter the data by
        :type sync_tables: bool
        :param sync_tables: sync table start/end by timestamp (relevant for running directly after DB table updates)
        :return: dictionary of statistics
        """

        RESET_THRESHOLD = 1000
        
        def sync_data_and_statistics_tables(data_table, statistics_table, gw_list):
            """
            sync data and statistics table by timestamps
            """
            def get_first_stat_timestamp(statistics_table):
                first_stat_timestamp = 0
                for gw in statistics_table['gatewayId'].unique():
                    gw_first_stat_timestamp = np.sort(statistics_table[statistics_table['gatewayId'] == gw]['timestamp'])[0]
                    if gw_first_stat_timestamp > first_stat_timestamp:
                        first_stat_timestamp = gw_first_stat_timestamp
                return first_stat_timestamp
            
            # 'sync' statistics table and data table latest timestamps:
            max_data_timestamp = data_table['timestamp'].max()
            max_stats_timestamp = statistics_table['timestamp'].max()
            last_synced = min(max_data_timestamp, max_stats_timestamp)
            # truncate end of data tables to timeslot
            data_table = data_table[data_table['timestamp']<= last_synced]
            statistics_table = statistics_table[statistics_table['timestamp'] <= last_synced]
            
            first_stat_timestamp = get_first_stat_timestamp(statistics_table)
            stat_start_timestamp = first_stat_timestamp - (datetime.timedelta(seconds = 60).total_seconds() * 1000)
            data_start_timestamp = data_table['timestamp'].min()
            while data_start_timestamp > stat_start_timestamp:
                statistics_table = statistics_table[statistics_table['timestamp']>first_stat_timestamp]
                first_stat_timestamp = get_first_stat_timestamp(statistics_table)
                stat_start_timestamp = first_stat_timestamp - (datetime.timedelta(seconds = 60).total_seconds() * 1000)

            return data_table, statistics_table
                

        
        def get_stats_for_continuous_sequence_ids(sequence):
            """
            generate statistics for contiuous sequence IDs
            :type sequence: list
            :param sequence: array of sequence ID
            :rtype: dict
            :return: dictionary of statistics
            """
            stats = {}

            # remove duplicates
            sequence = np.unique(sequence)
            # sort by descending order
            sequence = -np.sort(-sequence)

            s_max = np.max(sequence)
            s_min = np.min(sequence)
            s_rec = len(np.unique(sequence)) # TODO - Compare with len(sequence) / num of coupled packets from table
            s_expected = (s_max - s_min)+1
            stats['maxSequenceId'] = s_max
            stats['minSequenceId'] = s_min
            stats['receivedSequenceIds'] = s_rec
            stats['expectedSequenceIds'] = s_expected
            return stats

        def process_sequences(df_array, num_mgmt_packets=0):
            """
            generate statistics for (normalized by duration) for array of sequence IDs (compensated for GW Reboots)
            :type df_array: list of DataFrames
            :param df_array: list of DataFrames with continuous sequence IDs
            :type num_mgmt_packets: int
            :param num_mgmt_packets: number of management packets
            :rtype: dict
            :return: dictionary of statistics
            """
            total_stats = {}
            total_received_packets = num_mgmt_packets
            total_expected_packets = 0
            for df in df_array:
                stats = get_stats_for_continuous_sequence_ids(df['sequenceId'])
                total_received_packets += stats['receivedSequenceIds']
                total_expected_packets += stats['expectedSequenceIds']
            if total_expected_packets == 0:
                breakpoint()
            loss_percentage = (1 - (total_received_packets / total_expected_packets) )* 100
            total_stats['totalManagementPackets'] = num_mgmt_packets
            total_stats['totalDataPackets'] = total_received_packets - num_mgmt_packets
            total_stats['totalReceivedPackets'] = total_received_packets
            total_stats['totalExpectedPackets'] = total_expected_packets
            total_stats['lossPercentage'] = round(loss_percentage, 3)
            total_stats['numResets'] = len(df_array)-1
            return total_stats

        results = {}
        if isinstance(packets_table, pd.DataFrame):
            data = packets_table
        else:
            data = self.get_sequence_id_data(packets_table, start_time, end_time, gateways_list).sort_values(by='timestamp', ascending=False)
        gw_list = data['gatewayId'].unique()
        if gw_list is None:
            return None
        if isinstance(statistics_table, pd.DataFrame):
            statistics_data = statistics_table
        else:
            statistics_data = self.get_statistics_data(statistics_table, start_time, end_time, gw_list)
        if sync_tables:
            data, statistics_data = sync_data_and_statistics_tables(data, statistics_data, gw_list)

        for gw in gw_list:
            gw_name = data[data['gatewayId'] == gw]['gatewayName'].iloc[0]
            num_mgmt_packets = statistics_data[statistics_data['gatewayId'] == gw]['managementPktCount'].sum()
            gw_results = []
            gw_df = data[data['gatewayId'] == gw].reset_index()
            gw_df['diff'] = gw_df['sequenceId'].diff()
            gw_df['reset'] = np.where(gw_df['diff']>RESET_THRESHOLD, True, False)
            gw_resets = gw_df[gw_df['reset'] == True]
            sequences = np.array_split(gw_df, gw_resets.index)
            gw_results = process_sequences(sequences, num_mgmt_packets)
            gw_results.update({'gwName': gw_name})
            results[gw] = gw_results
        return results



    def get_amount_of_unique_tags_per_data_path(self, table, start_time, end_time, gateways_list=None,
                                                tags_to_ignore=None, bridges_list=None):
        """
        does an SQL query of the packet data table between specified timestamps
        returns amount of unique externalIds per data path (bridge->gw) in the given timeframe
        :type table: str
        :param table: name of data table
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :type tags_to_ignore: list
        :param tags_to_ignore: list of tags to ignore in the query (will not be counted)
        :type gateways_list: list
        :param gateways_list: list of gateways to filter the data by
        :type bridges_list: list
        :param bridges_list: list of bridges to filter the data by
        :rtype: pandas DataFrame
        :return: dataframe of data from table
        """
        if bridges_list is not None:
            if len(bridges_list) == 1:
                bridges_list = list(bridges_list) + [""]
            bridges_list = tuple(bridges_list)
        if gateways_list is not None:
            if len(gateways_list) == 1:
                gateways_list = list(gateways_list) + [""]
            gateways_list = tuple(gateways_list)
        if tags_to_ignore is not None:
            if len(tags_to_ignore) == 1:
                tags_to_ignore = list(tags_to_ignore) + [""]
            tags_to_ignore = tuple(tags_to_ignore)
        # adding search by date to improve search time (most tables in data bricks partition is by date)
        query_start_datetime = convert_timestamp_to_datetime(str(start_time))
        query_end_datetime = convert_timestamp_to_datetime(str(end_time))
        start_datetime = query_start_datetime - datetime.timedelta(hours=36)
        end_datetime = query_end_datetime + datetime.timedelta(hours=36)
        start_date = datetime.datetime.strftime(start_datetime, '%Y%m%d')
        end_date = datetime.datetime.strftime(end_datetime, '%Y%m%d')
        sql_method = f"""
        select gatewayId, bridgeId, count(distinct externalId)
        from {table}
        where date between {start_date} and {end_date} 
        and timestamp between {start_time} and {end_time} 
        """
        if tags_to_ignore is not None and tags_to_ignore != ():
            sql_method = sql_method + f"""and externalId not in {tags_to_ignore}
        """
        if bridges_list is not None and bridges_list != ():
            sql_method = sql_method + f"""and bridgeId in {bridges_list}
        """
        if gateways_list is not None and gateways_list != ():
            sql_method = sql_method + f"""and gatewayId in {gateways_list}
        """
        sql_method = sql_method + f"""group by gatewayId, bridgeId   order by gatewayId, bridgeId"""
        if self.spark is not None:
            debug_print('Running SQL query...', center=True)
            debug_print(sql_method)
            query_data = self.spark.sql(sql_method)
            query_data = query_data.toPandas()
            return query_data
        else:
            raise EnvironmentError("SQL query can only run in databricks")

    def get_amount_of_unique_tags_per_brg(self, table, start_time, end_time, tags_to_ignore=None, bridges_list=None):
        """
        does an SQL query of the packet data table between specified timestamps
        returns amount of unique externalIds per data path (bridge->gw) in the given timeframe
        :type table: str
        :param table: name of data table
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :type tags_to_ignore: list
        :param tags_to_ignore: list of tags to ignore in the query (will not be counted)
        :type gateways_list: list
        :param gateways_list: list of gateways to filter the data by
        :type bridges_list: list
        :param bridges_list: list of bridges to filter the data by
        :rtype: pandas DataFrame
        :return: dataframe of data from table
        """
        if bridges_list is not None:
            if len(bridges_list) == 1:
                bridges_list = list(bridges_list) + [""]
            bridges_list = tuple(bridges_list)
        if tags_to_ignore is not None:
            if len(tags_to_ignore) == 1:
                tags_to_ignore = list(tags_to_ignore) + [""]
            tags_to_ignore = tuple(tags_to_ignore)
        # adding search by date to improve search time (most tables in data bricks partition is by date)
        query_start_datetime = convert_timestamp_to_datetime(str(start_time))
        query_end_datetime = convert_timestamp_to_datetime(str(end_time))
        start_datetime = query_start_datetime - datetime.timedelta(hours=36)
        end_datetime = query_end_datetime + datetime.timedelta(hours=36)
        start_date = datetime.datetime.strftime(start_datetime, '%Y%m%d')
        end_date = datetime.datetime.strftime(end_datetime, '%Y%m%d')
        sql_method = f"""
        select gatewayId, bridgeId, count(distinct externalId)
        from {table}
        where date between {start_date} and {end_date} 
        and timestamp between {start_time} and {end_time} 
        """
        if tags_to_ignore is not None and tags_to_ignore != ():
            sql_method = sql_method + f"""and externalId not in {tags_to_ignore}
        """
        if bridges_list is not None and bridges_list != ():
            sql_method = sql_method + f"""and bridgeId in {bridges_list}
        """
        sql_method = sql_method + f"""group by bridgeId   order by bridgeId"""
        if self.spark is not None:
            debug_print('Running SQL query...', center=True)
            debug_print(sql_method)
            query_data = self.spark.sql(sql_method)
            query_data = query_data.toPandas()
            return query_data
        else:
            raise EnvironmentError("SQL query can only run in databricks")

    def get_amount_of_unique_tags_per_gw(self, table, start_time, end_time, gateways_list=None, tags_to_ignore=None):
        """
        does an SQL query of the packet data table between specified timestamps
        returns amount of unique externalIds per data path (bridge->gw) in the given timeframe
        :type table: str
        :param table: name of data table
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :type tags_to_ignore: list
        :param tags_to_ignore: list of tags to ignore in the query (will not be counted)
        :type gateways_list: list
        :param gateways_list: list of gateways to filter the data by
        :rtype: pandas DataFrame
        :return: dataframe of data from table
        """

        if gateways_list is not None:
            if len(gateways_list) == 1:
                gateways_list = list(gateways_list) + [""]
            gateways_list = tuple(gateways_list)
        if tags_to_ignore is not None:
            if len(tags_to_ignore) == 1:
                tags_to_ignore = list(tags_to_ignore) + [""]
            tags_to_ignore = tuple(tags_to_ignore)
        # adding search by date to improve search time (most tables in data bricks partition is by date)
        query_start_datetime = convert_timestamp_to_datetime(str(start_time))
        query_end_datetime = convert_timestamp_to_datetime(str(end_time))
        start_datetime = query_start_datetime - datetime.timedelta(hours=36)
        end_datetime = query_end_datetime + datetime.timedelta(hours=36)
        start_date = datetime.datetime.strftime(start_datetime, '%Y%m%d')
        end_date = datetime.datetime.strftime(end_datetime, '%Y%m%d')
        sql_method = f"""
        select gatewayId, count(distinct externalId)
        from {table}
        where date between {start_date} and {end_date} 
        and timestamp between {start_time} and {end_time} 
        """
        if tags_to_ignore is not None and tags_to_ignore != ():
            sql_method = sql_method + f"""and externalId not in {tags_to_ignore}
        """
        if gateways_list is not None and gateways_list != ():
            sql_method = sql_method + f"""and gatewayId in {gateways_list}
        """
        sql_method = sql_method + f"""group by gatewayId   order by gatewayId"""
        if self.spark is not None:
            debug_print('Running SQL query...', center=True)
            debug_print(sql_method)
            query_data = self.spark.sql(sql_method)
            query_data = query_data.toPandas()
            return query_data
        else:
            raise EnvironmentError("SQL query can only run in databricks")

    def get_last_data(self, table, mins=60, tags_list=None, bridges_list=None, gateways_list=None):
        """
        function querys back specified minutes from data table (counts back from last entry in table)
        :type table: str
        :param table: databricks data table name
        :type mins: int
        :param mins: minutes to query back, defaults to 1 hour
        :param tags_list: tags list
        :type tags_list: list
        :param tags_list: list of tags to filter from the data
        :type bridges_list: list
        :param bridges_list: list of bridger to filter from the data
        :type gateways_list: list
        :param gateways_list: list of gateways to filter from the data
        :rtype: pandas DataFrame
        :return: dataframe of data from table
        """
        end_timestamp = self.get_last_entry_timestamp(table)
        end_datetime = mstimestamp_to_timezone(end_timestamp)
        start_timestamp = end_timestamp - (mins * 60 * 1000)
        start_datetime = mstimestamp_to_timezone(start_timestamp)
        debug_print(f'Getting last {mins} mins from {table}')
        debug_print(f'Last entry at {end_timestamp}')
        debug_print(f'querying {start_datetime} -> {end_datetime}')
        return self.get_seen_tags(table, start_timestamp, end_timestamp, tags_list, bridges_list, gateways_list)

    def get_last_entry_timestamp(self, table):
        """
        function gets name of data table and returns the timestamp of the last entry (in milliseconds)
        :type table: str
        :param table: name of table
        """
        today = datetime.datetime.now()
        yesterday = today-datetime.timedelta(days=1)
        today_date = datetime.datetime.strftime(today, '%Y%m%d')
        yesterday_date = datetime.datetime.strftime(yesterday, '%Y%m%d')

        # Try to get last timestamp from last day
        sql_method = f"""
        select MAX (timestamp)
        from {table}
        where date between {yesterday_date} and {today_date} 
        """
        if self.spark is not None:
            query_data = self.spark.sql(sql_method)
            query_data = query_data.toPandas()
            last_entry_ts = query_data.iloc[0][0]
            if not np.isnan(last_entry_ts):
                return query_data.iloc[0][0]
            else:
                # Query all packets in table
                sql_method = ''.join(sql_method.split('\n')[:-2])
                query_data = self.spark.sql(sql_method)
                query_data = query_data.toPandas()
                last_entry_ts = query_data.iloc[0][0]
                if not np.isnan(last_entry_ts):
                    return query_data.iloc[0][0]
                else:
                    raise WiliotTableError(f'Cannot get last entry, no entries in data table {table}')
        else:
            raise EnvironmentError("Unable to detect dbutils function")

    def wait_for_data(self, table, requested_timestamp, timeout_mins=70):
    # TODO - implement date search
        """
        function waits for data timed at requested_timestamp (or later) to appear in the data table.
        :type table: str
        :param table: data table name
        :type requested_timestamp: int
        :param requested_timestamp: timestamp to wait to appear in the table
        """
        last_entry_timestamp = self.get_last_entry_timestamp(table)
        table_ready = last_entry_timestamp >= requested_timestamp
        start = current_timestamp()
        while not table_ready:
            if (current_timestamp() - start) >= 1000 * 60 * timeout_mins:
                debug_print(f"""{timeout_mins} minute timeout reached! Table still does not have the requested data
                Run test again or get raw data with wait_for_data=False""")
                exit_notebook(f'Wait for table timeout reached after {timeout_mins} minutes')
            debug_print(
                f'Waiting for table to get data. \n'
                f'Requested timestamp {requested_timestamp} | ({mstimestamp_to_timezone(requested_timestamp)}) \n'
                f'Table latest timestamp {last_entry_timestamp} | ({mstimestamp_to_timezone(last_entry_timestamp)})')
            mins_behind = (requested_timestamp - last_entry_timestamp) / 60000
            debug_print(f'Data table {mins_behind} minutes behind , trying again in 30 seconds', center=True)
            sleep(30)
            last_entry_timestamp = self.get_last_entry_timestamp(table)
            table_ready = last_entry_timestamp >= requested_timestamp
        debug_print(
            f'Table is updated. \n'
            f'Requested timestamp {requested_timestamp}, table latest timestamp {last_entry_timestamp} \n'
            f'Table latest timestamp {last_entry_timestamp} | ({mstimestamp_to_timezone(last_entry_timestamp)})')
        mins_ahead = (last_entry_timestamp - requested_timestamp) / 60000
        debug_print(f'Data table {mins_ahead} minutes ahead', center=True)

    def get_seen_edge_devices_from_packets(self, table, start_time, end_time):
        """
        does an SQL query of the packet data table between specified timestamps
        returns dictionary of seen gateways and bridges from slice of packet table
        :type table: str
        :param table: name of data table
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :rtype: pandas DataFrame
        :return: dataframe of data from table
        """
        # adding search by date to improve search time (most tables in data bricks partition is by date)
        query_start_datetime = convert_timestamp_to_datetime(str(start_time))
        query_end_datetime = convert_timestamp_to_datetime(str(end_time))
        start_datetime = query_start_datetime - datetime.timedelta(hours=24)
        end_datetime = query_end_datetime + datetime.timedelta(hours=24)
        start_date = datetime.datetime.strftime(start_datetime, '%Y%m%d')
        end_date = datetime.datetime.strftime(end_datetime, '%Y%m%d')
        sql_method = f"""
        select gatewayId, bridgeId, gatewayName, max(timestamp) as timestamp
        from {table}
        where date between {start_date} and {end_date} 
        and timestamp between {start_time} and {end_time}
        group by gatewayId, gatewayName, bridgeId 
        """
        if self.spark is not None:
            debug_print('Running SQL query...', center=True)
            debug_print(sql_method)
            query_data = self.spark.sql(sql_method).toPandas()
            query_data['gatewayId'] = query_data['gatewayId'].str.upper()
            query_data['bridgeId'] = query_data['bridgeId'].str.upper()
            return query_data
        else:
            raise EnvironmentError("SQL query can only run in databricks")


def process_tagstats(params, tag_stats_df, working_directory, test_no):
    """
        function gets tag stats dataframe and creates downloadable cdf graph for specified parameters
        :type params: list
        :param params: tag stats metrics to create graphs for
        :type tag_stats_df: pandas DataFrame
        :param tag_stats_df: tag stats DataFrame (can be created using get_tagstats_from_test)
        :type working_directory: string
        :param working_directory: directory to save the graphs HTML files
        :type test_no: string
        :param test_no: test number
    """
    if params is None:
        debug_print('No parameters specified. Creating graphs for all possible parameters.')
        params = ['num_packets', 'num_cycles', 'sprinkler_counter_mean', 'sprinkler_counter_std',
                  'sprinkler_counter_min', 'sprinkler_counter_max', 'tbp_mean', 'tbp_std', 'tbp_min', 'tbp_max',
                  'tbp_num_vals', 'per_mean', 'per_std', 'rssi_mean', 'rssi_std', 'rssi_min', 'rssi_max', 'ttfp',
                  'ttfp_seconds', 'end_time', 'duration', 'rx_rate_normalized', 'rx_rate', 'charge_time_min',
                  'charge_time_max', 'packet_counter_min', 'packet_counter_max', 'packet_counter_first',
                  'estimated_packet_counter_resets', 'estimated_total_packet_rate', 'estimated_total_per',
                  'externalId']
    tests_list = tag_stats_df['testId'].unique()
    for param in params:
        debug_print(f"""*****************************{param}*****************************""")
        graph = pd.DataFrame()
        for test in tests_list:
            df = tag_stats_df[tag_stats_df.testId == test]
            test_curveset = cdf_dataframe(df, param, test)
            graph = pd.concat([test_curveset, graph])
            debug_print(f"""test {test} added to graph""")
        fig = px.scatter(graph, x=param, y='TagCount', color=['testName'])
        fig.update_layout(title=param)
        fig.update_yaxes()
        if param == 'ttfp_seconds':
            fig.update_xaxes(title_text='Time to first packet [seconds]')
        if param == 'estimated_total_packet_rate':
            fig.update_xaxes(title_text='Estimated total packet rate [counters/second]')
        filepath = f"""{working_directory}{test_no}_{param}_graph.html"""
        open(filepath, 'a').close()
        fig.write_html(filepath)
        # debug_print(filepath)
        # p = plot(fig, output_type='div')
        # display_html(p)
        create_download_link(filepath, file_name=f"""{param} graph""")

def file_exists(path):
    """
    function checks if file exists
    :type path: str
    :param path: path to file
    :rtype: bool
    :return: file exists
    """
    path = path.replace('/dbfs', 'dbfs:')
    try:
        db_utils().fs.head(path)
        return True
    except Exception:
        try:
            return exists(path)
        except Exception:
            return False

def get_tagstats_from_test(tests_config_df=None, tests_config_csv_path=None, rawdata_df=None,
                           rawdata_csv_path=None, tagdeployment_df=None, tagdeployment_csv_path=None):
    """
        function gets test config df/csv and rawdata (from datatables) df/csv as input
        the function processes the data and created a new dataframe containing tag statistics for each test.
        :type tests_config_df: pandas DataFrame
        :param tests_config_df: dataframe of test config
        :type tests_config_csv_path: string
        :param tests_config_csv_path: path to test config csv
        :type rawdata_df: pandas DataFrame
        :param rawdata_df: dataframe of raw data
        :type rawdata_csv_path: string
        :param rawdata_csv_path: path to raw data csv
        :type tagdeployment_df: pandas DataFrame
        :param tagdeployment_df:dataframe of tag deployment data
        :type tagdeployment_csv_path: string
        :param tagdeployment_csv_path: path to tag deployment data csv
    """
    if tests_config_csv_path is not None and not tests_config_df:
        tests_config_df = pd.read_csv(tests_config_csv_path)
    if rawdata_csv_path is not None and not rawdata_df:
        rawdata_df = pd.read_csv(rawdata_csv_path)
    if tagdeployment_csv_path is not None:
        tagdeployment_df = pd.read_csv(tagdeployment_csv_path)
    testid_list = tests_config_df['testId'].unique()
    master_tag_stats = pd.DataFrame()
    for testid in testid_list:
        if  testid not in rawdata_df.testId.unique():
            debug_print(f"""0 data for testId {testid} in raw data! please check tables/test config""")
            continue
        currenttest_df = rawdata_df[rawdata_df.testId == testid].copy()
        start_time = tests_config_df[tests_config_df.testId == testid].get('startTimestamp').item()
        debug_print(f"""***** Processing {testid} data. Test Timestart = {start_time}""")
        currenttest_packets = analyze_packets_data(packets_df=currenttest_df, test_time_start=start_time)
        tag_stats_dict = currenttest_packets.get_group_statistics(group_by_col='externalId')
        tag_stats_df = pd.DataFrame.from_dict(tag_stats_dict, orient='index')
        tag_stats_df['testId'] = testid
        master_tag_stats = pd.concat([master_tag_stats, tag_stats_df])
    if tagdeployment_df is not None:
        debug_print('***********Tag deployment information merged with tag stats from test***********')
        master_tag_stats = pd.merge(master_tag_stats, tagdeployment_df, how='outer', on='externalId'). \
            sort_values('testId').set_index('externalId')
    master_tag_stats.insert(18, 'ttfp_seconds', master_tag_stats.ttfp / 1000)
    return master_tag_stats

def massagedata(dataset, param):
    curveset = pd.DataFrame(columns=[param, 'TagCount'])
    count = 0
    reftme_stmp = None
    for index, row in dataset.iterrows():
        if pd.isna(row[param]):
            continue
        if reftme_stmp is None:
            reftme_stmp = row[param]
        if reftme_stmp == row[param]:
            count += 1
        else:
            curveset = curveset.append({param: reftme_stmp, 'TagCount': count}, ignore_index=True)
            reftme_stmp = row[param]
            count += 1
    curveset = curveset.append({param: reftme_stmp, 'TagCount': count}, ignore_index=True)
    return curveset

def display_html(html):
    """
    Use databricks displayHTML from an external package
    :type html: string
    :param html: html document to display
    """
    for frame in inspect.getouterframes(inspect.currentframe()):
        global_names = set(frame.frame.f_globals)
        # Use multiple functions to reduce risk of mismatch
        if all(v in global_names for v in ["displayHTML", "display", "spark"]):
            return frame.frame.f_globals["displayHTML"](html)
    raise EnvironmentError("Unable to detect displayHTML function")

def display_print(todisplay, console_only=False, **kwargs):
    """
    Use databricks display func from an external package.
    uses tabulate library to display dataframe in case running locally
    :type todisplay: pandas DataFrame
    :param todisplay: variable to display
    :type console_only: bool
    :param console_only: if true, prints the data table to console (even if running in DB notebook) using tabulate
    """
    if not console_only:
        for frame in inspect.getouterframes(inspect.currentframe()):
            # call dbutils display
            global_names = set(frame.frame.f_globals)
            if all(v in global_names for v in ["display"]):
                try:
                    return frame.frame.f_globals["display"](todisplay)
                except ValueError:
                    debug_print('ValueError when reading DataFrame! Trying to print in console', center=True)
                except KeyError:
                    debug_print('KeyError when reading DataFrame! Trying to print in console', center=True)
                except TypeError:
                    debug_print('TypeError when reading DataFrame! Trying to print in console', center=True)
    if isinstance(todisplay, pd.DataFrame) or console_only:
        debug_print('\n' + tabulate.tabulate(todisplay, **kwargs))
        return None
    raise EnvironmentError("Unable to detect Display function")

def db_utils_rm(dbfs_path):
    """
    helper function to remove files/folders from dbfs
    """
    dbfs_path = dbfs_path.replace('/dbfs', 'dbfs:')
    return db_utils().fs.rm(dbfs_path)

def db_utils():
    """
    Use databricks dbutils from an external package
    for example: to use dbutils.fs.head() - use db_utils().fs.head()
    """
    for frame in inspect.getouterframes(inspect.currentframe()):
        global_names = set(frame.frame.f_globals)
        # Use multiple functions to reduce risk of mismatch
        if all(v in global_names for v in ["dbutils"]):
            return frame.frame.f_globals["dbutils"]()
    raise EnvironmentError("Unable to detect dbutils function")

def get_secret(scope, key):
    """
    get databrickst secret, return None if not running in databricks
    :param scope: secret scope
    :param key: secret key
    """
    try:
        secret = db_utils().secrets.get(scope=scope, key=key)
        return secret
    except EnvironmentError:
        raise(EnvironmentError('Cannot get secret when not running in databricks!'))


def create_download_link(dbfs_path, file_name='file'):
    """
    accepts path to dbfs file, and creates a download link to the file in the notebook
    the function only works with files saved in /dbfs/FileStore
    :type dbfs_path: string
    :param dbfs_path: path to dbfs path (accepts either spark API or file API format)
    :type file_name: string
    :param file_name: name of file (this will be the name of the link created)
    """
    if is_databricks():
        if not ('/dbfs/FileStore' in dbfs_path or 'dbfs:/FileStore' in dbfs_path):
            raise ValueError('the path must start with /dbfs/FileStore or dbfs:/FileStore!')
        dbfs_path = dbfs_path.replace('/dbfs/FileStore', '/files')
        dbfs_path = dbfs_path.replace('dbfs:/FileStore', '/files')
        display_html(f"""<a href="{dbfs_path}" download>Download {file_name} </a>""")
    debug_print(f"""File available at {dbfs_path}""")

def cdf_dataframe(dataset, param, dataset_name, groupby=None, comp=False):
    """
    function accepts dataframe as input the function returns dataframe of CDF
    complementary cumulative distribution function (of tag count) according to param,
    and labels it by dataset_name
    :type dataset: pandas DataFrame
    :param dataset: dataset to create CDF
    :type param: string
    :param param: parameter to create the CDF
    :type dataset_name: string
    :param dataset_name: name to add to the result
    :type groupby: str
    :param groupby: added parameter (in dataset) to group results by
    :type comp: bool
    :param comp: whether to create CDF (comp = False) or CCDF (comp = True)
    """
    curveset = pd.DataFrame(columns=[param, 'TagCount', groupby])
    dataset = dataset.sort_values(param, ascending=True)
    if groupby is None:
        iterations = [None]
    else:
        iterations = dataset[groupby].unique()
    for group in iterations:
        group_curveset = pd.DataFrame(columns=[param, 'TagCount', groupby])
        if group is not None:
            group_dataset = dataset[dataset[groupby] == group]
        else:
            group_dataset = dataset
        values = group_dataset[param].unique()
        group_curveset[param] = values
        if group is not None:
            group_curveset[groupby] = group
        for index, value in group_curveset[param].items():
            biggerthan = group_dataset[group_dataset[param] > value][param].count()
            smallerorequals = group_dataset[group_dataset[param] <= value][param].count()
            if comp:
                group_curveset.at[index, 'TagCount'] = biggerthan
            else:
                group_curveset.at[index, 'TagCount'] = smallerorequals
        if comp:
            group_curveset['testName'] = dataset_name + '_CCDF'
        else:
            group_curveset['testName'] = dataset_name + '_CDF'
        curveset = pd.concat([curveset, group_curveset])

    return curveset

def powerset(iterable):
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))

def analyze_packets_data(packets_df=None, csv_path=None, test_time_start=None, test_time_conf_received=None,
                         no_duplications=True):

    """
    gets dataframe of packets / csv of packets, returns NetworkPacketList of packets
    :type csv_path: string
    :param csv_path: path to the packets csv file
    :type packets_df: dataframe
    :param packets_df: packets dataframe
    :type test_time_start: timestamp (int)
    :param test_time_start: UTC millisecond timestamp of the beginning of the test
    :type test_time_conf_received: timestamp (int)
    :param test_time_conf_received: UTC millisecond timestamp of the time the test configuration was received
    :type no_duplications: bool
    :param no_duplications: if False, will keep duplications of same rawPacket
    :return: NetworkPacketList created from .csv located in path
    """
    if test_time_start is None:
        debug_print('Test time start was not passed, some metrics will not be calculated!')
    else:
        debug_print(f'Test time start {test_time_start} | ')
    if csv_path is None and packets_df is None:
        debug_print("no data was given!()")
        return None
    if csv_path is not None:
        packets_df = pd.read_csv(csv_path)

    debug_print('Preprocessing raw data...')

    # remove duplicate rows
    dup_rows = packets_df.shape[0]
    packets_df.sort_values(by='timestamp', ascending=True, ignore_index=True, inplace=True)
    if no_duplications:
        packets_df.drop_duplicates('rawPacket', ignore_index=True, inplace=True)
        dup_rows = dup_rows - packets_df.shape[0]
        if dup_rows > 0:
            debug_print(f'{dup_rows} duplicate rows removed')

    # flag bugged tags packets
    flag_bugged_tags(packets_df)

    debug_print('Creating NetworkPacketList from DataFrame...')
    packets_list = NetworkPacketList(list_custom_data={'test_time_start': test_time_start})
    if test_time_conf_received is not None:
        packets_list.list_custom_data['test_time_conf_received'] = test_time_conf_received
    reconstructed_packet_df = network_data_preparation(packets_df=packets_df)

    packets_list = packets_list.import_packet_df(import_all=True, packet_df=reconstructed_packet_df)
    return packets_list

def create_zip(dir_path):
    """
    gets path to FileStore directory and creates a download link to a zip file of the files in given directory
    zip file will be saved inside the dir_path directory
    :type dir_path: string
    :param dir_path: path to FileStore directory to zip
    """
    if dir_path[-1] == '/':
        dir_path = dir_path[:-1]
    dir_name = dir_path.rsplit('/', 1)[-1]
    if file_exists(dir_path + '/' + dir_name + '.zip'):
        db_utils().fs.rm(dir_path.replace('/dbfs', 'dbfs:') + '/' + dir_name + '.zip')
    zip_tmp_path = shutil.make_archive('/tmp/zips/tmpzip', 'zip', dir_path)
    zip_download_path = dir_path.replace('/dbfs', 'dbfs:') + '/' + dir_name + '.zip'
    db_utils().fs.mv(f'file:{zip_tmp_path}', zip_download_path)
    create_download_link(zip_download_path, f'{dir_name}.zip')

def exit_notebook(message):
    """
    closes notebook and prints out error message
    :type message: str
    :param message: error message
    """
    debug_print(message, center=True)
    try:
        db_utils().notebook.exit(message)
    except EnvironmentError as e:
        raise Exception(message)

def save_df(df, path, name=None, withindex=False, silent=False):
    """
    saves DataFrame to path, displays the DataFrame and creates a download link
    :type df: pandas Dataframe
    :param df: dataframe to save
    :type path: str
    :param path: path to save can be entered either as filename to save or directory to save in
    :type name: str
    :param name: name of dataframe (this displays in the download link and as a header before the link
    :type withindex: bool
    :param withindex: flag to choose if to export the dataframe with/without index
    :type silent: bool
    :param silent: if true does not generate HTML link
    """
    # save as dataframe.csv if no filename given
    if path[-1] == '/':
        path = path + 'dataframe'
    if path[-4:] != '.csv':
        path = path + '.csv'
    if name is None:
        name = path.split()[-1]
    if is_databricks():
        path = path.replace(' ', '_')
    debug_print(f'Saving DataFrame at {path}')
    if not is_databricks():
        df.to_csv(path, index=withindex)
        return True
    try:
        df.to_csv(path, index=withindex)
    except OSError:
        mkdirs_path = path.replace('/dbfs', 'dbfs:').rsplit('/', 1)[0]
        db_utils().fs.mkdirs(mkdirs_path)
        df.to_csv(path, index=withindex)
    if not silent:
        if is_databricks():
            display_html(f'<h1>{name}</h1>')
            create_download_link(dbfs_path=path, file_name=name)
            display_print(pd.read_csv(path), headers="keys")
        else:
            debug_print(f'{name} available at {path}')

def initialize_logger(working_directory=None):
    """
    initializes the logger to print to log and to logfile, which by default is named by the current timestamp (ms)
    when calling the function.
    :param working_directory: working directory to save logfile (in case running locally)
    :type working_directory: str
    :return: logger fileHandler filename
    """
    logger = logging.getLogger()
    for handler in logger.handlers:
        try:
            # extract current filename from log
            filename = handler.baseFilename.split('\\')[-1].split('.')[0]
            debug_print(f'Logger already initialized! passing logfile {filename}')
            return filename
        except Exception:
            pass
    logger_filename = int(current_timestamp())
    if is_databricks():
        db_utils().fs.put(f'file:/databricks/driver/{logger_filename}.log', '', overwrite=True)
        logging.getLogger("py4j.java_gateway").setLevel(logging.ERROR)
        logging.basicConfig(level=logging.DEBUG,
                            format='%(asctime)s | %(levelname)s | %(message)s',
                            handlers=[
                                logging.FileHandler(f'{logger_filename}.log', 'a'),
                                # logging.handlers.RotatingFileHandler
                                logging.StreamHandler()
                            ])
        logging.getLogger().handlers[0].setLevel(logging.DEBUG)
        logging.getLogger().handlers[1].setLevel(logging.INFO)
        debug_print(f'logger initialized at {logger_filename}', center=True)
        debug_print(f'logfile located at file:/databricks/driver/{logger_filename}.log')
    else:
        if working_directory is None:
            working_directory = os.path.join(user_data_dir(), 'wiliot', 'deployment_tools')
        if not os.path.exists(working_directory):
            os.makedirs(working_directory)
        logging.basicConfig(level=logging.DEBUG,
                            format='%(asctime)s | %(levelname)s | %(message)s',
                            handlers=[
                                logging.FileHandler(f'{working_directory}/{logger_filename}.log', 'a'),
                                # logging.handlers.RotatingFileHandler
                                logging.StreamHandler()
                            ])
        # filter stream to show info and up
        logging.getLogger().handlers[0].setLevel(logging.DEBUG)
        logging.getLogger().handlers[1].setLevel(logging.INFO)
        debug_print(f'logger initialized at {logger_filename}', center=True)
        debug_print(f'logfile located at {working_directory}/{logger_filename}.log')
    logging.getLogger().setLevel(logging.DEBUG)
    return logger_filename

def create_logfile(logger_filename, working_directory, copy_filename):
    """
        function copies the logfile for recent run to the test working directory and creates link for download
        :type logger_filename: str
        :param logger_filename: filename of log file
        :type working_directory: str
        :param working_directory: directory to copy logfile to
        :type copy_filename: str
        :param copy_filename: name of logfile when copied
    """
    if not is_databricks():
        debug_print('Cannot create logfile, logfile available already!')
    debug_print('Creating Logfile...', center=True)
    debug_print(f'cp {logger_filename}.log {copy_filename}.log')
    subprocess.run(f'cp {logger_filename}.log {copy_filename}.log', shell=True)
    copy_directory = working_directory.replace('/dbfs', 'dbfs:')
    debug_print(f'copy directory {copy_directory}')
    db_utils().fs.mkdirs(f'{copy_directory}')
    debug_print(f'mkdirs {copy_directory}')
    db_utils().fs.cp(f'file:/databricks/driver/{copy_filename}.log', copy_directory)
    debug_print(f'cp file:/databricks/driver/{copy_filename}.log, copy_directory')
    create_download_link(working_directory + f'{copy_filename}.log', f'{copy_filename}.log')
    debug_print(f"create download link - {working_directory} + f'{copy_filename}.log', f'{copy_filename}.log'")

def get_packet_table_name(owner, env, platform=False):
    """
    function gets ownerId and environment and return the name of the packet table (in databricks)
    :type owner: str
    :param owner: ownerId
    :type env: str
    :param env: wiliot environment (prod/test/dev)
    :type platform: bool
    :param platform: wiliot platform
    :rtype: str
    :return: data table name
    """
    env = 'prod' if env is None else env
    if not platform:
        data_table = owner + '' + f'_packet_data_{env}'
    else:
        data_table = owner + f'_enriched_packets_data_{env}'
    if owner.isnumeric():
        data_table = '_' + data_table
    return data_table

def get_event_table_name(owner, env=None, platform=False):
    """
    function gets ownerId and environment and return the name of the event table (in databricks)
    :type owner: str
    :param owner: ownerId
    :type env: str
    :param env: wiliot environment (prod/test/dev)
    :type platform: bool
    :param platform: wiliot platform
    :rtype: str
    :return: data table name
    """
    env = 'prod' if env is None else env
    if not platform:
        event_table = owner + '' + f'_event_data_{env}'
    else:
        event_table = owner + '' + f'_assets_metrics_data_{env}'
    if owner.isnumeric():
        event_table = '_' + event_table
    return event_table

def get_heartbeat_table_name(env=None):
    """
    function gets environment and returns the name of the heartbeat table (in databricks)
    :type owner: str
    :param owner: ownerId
    :type env: str
    :param env: wiliot environment (prod/test/dev)
    :rtype: str
    :return: hearbeat table name
    """
    env = 'prod' if env is None else env
    hearbeat_table = f'_network_heartbeat_statistics_{env}'
    return hearbeat_table

def get_statistics_table_name(env=None):
    """
    function gets environment and returns the name of the heartbeat table (in databricks)
    :type owner: str
    :param owner: ownerId
    :type env: str
    :param env: wiliot environment (prod/test/dev)
    :rtype: str
    :return: hearbeat table name
    """
    env = 'prod' if env is None else env
    hearbeat_table = f'_network_data_statistics_{env}'
    return hearbeat_table

def get_configuration_table_name(owner, env=None):
    """
    function gets ownerId and environment and return the name of the configuration table (in databricks)
    :type owner: str
    :param owner: ownerId
    :type env: str
    :param env: wiliot environment (prod/test/dev)
    :rtype: str
    :return: configuration table name
    """
    env = 'prod' if env is None else env
    configuration_table = owner + '' + f'_network_configuration_{env}'
    if owner.isnumeric():
        configuration_table = '_' + configuration_table
    return configuration_table

def process_graph(fig, graph_name, display_graph=True, save_graph=False, directory=None, pass_as_is=False):
    """
    function processes graph - displays and saves the graph (according to entered flags)
    :type fig: plotly Figure
    :param fig: graph to display
    :type graph_name: str
    :param graph_name: name of graph, and the filename by which it will be saved
    :type display_graph: bool
    :param display_graph: flag to choose if to display the graph in DataBricks Notebook
    :type save_graph: bool
    :param save_graph: flag to choose if to save the graph
    :type directory: str
    :param directory: directory to save graph
    :type pass_as_is: bool
    :param pass_as_is: if true the fig entered is already a plot
    """
    debug_print(f'Processing Graph - {graph_name}', center=True)
    if save_graph:
        if directory is None:
            exit_notebook('Need to supply directory to save graph!')
        if directory[-1] != '/':
            directory = directory + '/'
        filepath = f"""{directory}{graph_name}.html"""
        Path(directory).mkdir(parents=True, exist_ok=True)
        open(filepath, 'a').close()
        fig.write_html(filepath)
        create_download_link(filepath, file_name=graph_name)
    if display_graph:
        if is_databricks():
            if not pass_as_is:
                fig = plot(fig, output_type='div')
            display_html(fig)
        else:
            fig.show(renderer="browser")

def get_spark():
    try:
        from pyspark.sql import SparkSession
        return SparkSession.builder.getOrCreate()
    except NameError:
        raise NameError('Spark is not installed!')


def analyze_packets_data_per_brg(df, brg_list, time_start, test_time_conf_received=None):
    df = df.drop_duplicates(subset=["rawPacket", "bridgeId"])
    tag_stats_df = pd.DataFrame()
    for brg in brg_list:
        brg_df = df[df['bridgeId'] == brg]
        debug_print(f'Processing {brg_df.shape[0]} packets from {brg}')
        if brg_df.empty:
            continue
        df_tagstats = analyze_packets_data(packets_df=brg_df, test_time_start=time_start,
                                           test_time_conf_received=test_time_conf_received)
        tag_stats_dict = df_tagstats.get_group_statistics(group_by_col='externalId')
        tag_stats_df_tmp = pd.DataFrame.from_dict(tag_stats_dict, orient='index')
        tag_stats_df_tmp['bridgeId'] = brg
        tag_stats_df = pd.concat([tag_stats_df, tag_stats_df_tmp])
    if 'externalId' not in tag_stats_df.columns:
        tag_stats_df['externalId'] = tag_stats_df.index
    tag_stats_df = tag_stats_df.sort_values('rx_rate_normalized', ascending=False).drop_duplicates(
        'externalId').sort_index()
    return tag_stats_df


def flag_bugged_tags(rawdata_df):
    """
    function gets raw data dataframe, adds 'isBugged' column
    'isBugged' is true for each packet of a bugged tag, false otherwise
    :type rawdata_df: pandas DataFrame
    :param rawdata_df: raw packet data
    """
    rawdata_df['isBugged'] = None
    unique_tags = rawdata_df['externalId'].unique()
    for tag in unique_tags:
        tmp_df = rawdata_df.loc[rawdata_df['externalId'] == tag]
        tmp_df = tmp_df[['packet_counter', 'timestamp']]
        tmp_df = tmp_df.sort_values(by=['timestamp'])
        prev_packet_counter = 0
        cycles = 0
        orig_cycles = tmp_df['packet_counter'].unique()
        for timestamp in tmp_df['timestamp'].unique():
            # TODO - see when this function throws error
            try:
                packet_counter = tmp_df.loc[tmp_df['timestamp'] == timestamp, 'packet_counter'].unique().item()
            except Exception:
                continue
            if packet_counter + 256 * cycles < prev_packet_counter:
                cycles = cycles + 1
            prev_packet_counter = packet_counter + 256 * cycles
            tmp_df.loc[tmp_df['timestamp'] == timestamp, 'packet_counter'] = prev_packet_counter
        tmp_df2 = tmp_df.diff(axis=0)
        tmp_df2['rate'] = (1000 * tmp_df2['packet_counter'] / tmp_df2['timestamp'])
        max_rate = tmp_df2['rate'].max()
        is_bugged = False
        if max_rate > 6:
            is_bugged = True
            debug_print(f'{tag} is bugged! Flagging packets')
        rawdata_df.loc[rawdata_df['externalId'] == tag, 'isBugged'] = is_bugged


def parse_commastring(string):
    """
    parse string with comma or comma+space seperated values to list
    :type string: str
    :param string: input
    :rtype: list
    :return: list of values
    """
    if type(string) == float:
        try:
            if np.isnan(string):
                return None
        except Exception as e:
            pass
    if type(string) == list:
        return string
    if string is None:
        return list()
    cmd_list = ''.join(string.split()).split(',')
    return cmd_list


def parse_commastring_array(array):
    """
    parses Pandas array (DataFrame column) to list of all unique values
    :type array: Pandas array
    :param array: array
    :rtype: list
    :return: list of unique values
    """
    if len(array) == 1:
        return parse_commastring(array[0])
    result = list()
    for item in array:
        item_values = parse_commastring(item)
        if len(item_values) == 1:
            item_values
            if item_values[0] not in result:
                result.extend(item_values)
        else:
            for value in item_values:
                if value not in result:
                    result.extend([value])
    return result